# ============================================================================ #
#                                                                              #
#     Title   : Dataframe Cleaning                                             #
#     Purpose : Fetch columns from a given DataFrame using convenient syntax.  #
#                                                                              #
# ============================================================================ #


# ---------------------------------------------------------------------------- #
#                                                                              #
#     Overview                                                              ####
#                                                                              #
# ---------------------------------------------------------------------------- #


# ---------------------------------------------------------------------------- #
#  Description                                                              ####
# ---------------------------------------------------------------------------- #


"""
!!! note "Summary"
    The `columns` module is used to fetch columns from a given DataFrame using convenient syntax.
"""


# ---------------------------------------------------------------------------- #
#                                                                              #
#     Setup                                                                 ####
#                                                                              #
# ---------------------------------------------------------------------------- #


# ---------------------------------------------------------------------------- #
#  Imports                                                                  ####
# ---------------------------------------------------------------------------- #


# ## Python StdLib Imports ----
from typing import Literal, Optional, Union

# ## Python Third Party Imports ----
from pyspark.sql import DataFrame as psDataFrame
from toolbox_python.checkers import is_type
from toolbox_python.collection_types import str_collection, str_list
from typeguard import typechecked

# ## Local First Party Imports ----
from toolbox_pyspark.checks import (
    assert_columns_exists,
    warn_columns_missing,
)


# ---------------------------------------------------------------------------- #
#  Exports                                                                  ####
# ---------------------------------------------------------------------------- #


__all__: str_list = [
    "get_columns",
    "get_columns_by_likeness",
    "rename_columns",
    "reorder_columns",
    "delete_columns",
]


# ---------------------------------------------------------------------------- #
#                                                                              #
#     Functions                                                             ####
#                                                                              #
# ---------------------------------------------------------------------------- #


# ---------------------------------------------------------------------------- #
#  Selecting                                                                ####
# ---------------------------------------------------------------------------- #


@typechecked
def get_columns(
    dataframe: psDataFrame,
    columns: Optional[Union[str, str_collection]] = None,
) -> str_list:
    """
    !!! note "Summary"
        Get a list of column names from a DataFrame based on optional filter criteria.

    Params:
        dataframe (psDataFrame):
            The DataFrame from which to retrieve column names.
        columns (Optional[Union[str, str_collection]], optional):
            Optional filter criteria for selecting columns.<br>
            If a string is provided, it can be one of the following options:

            | Value | Description |
            |-------|-------------|
            | `#!py "all"` | Return all columns in the DataFrame.
            | `#!py "all_str"` | Return columns of string type.
            | `#!py "all_int"` | Return columns of integer type.
            | `#!py "all_numeric"` | Return columns of numeric types (integers and floats).
            | `#!py "all_datetime"` or `#!py "all_timestamp"` | Return columns of datetime or timestamp type.
            | `#!py "all_date"` | Return columns of date type.
            | Any other string | Return columns matching the provided exact column name.

            If a list or tuple of column names is provided, return only those columns.<br>
            Defaults to `#!py None` (which returns all columns).

    Raises:
        TypeError:
            If any of the inputs parsed to the parameters of this function are not the correct type. Uses the [`@typeguard.typechecked`](https://typeguard.readthedocs.io/en/stable/api.html#typeguard.typechecked) decorator.

    Returns:
        (str_list):
            The selected column names from the DataFrame.

    ???+ example "Examples"

        ```{.py .python linenums="1" title="Set up"}
        >>> # Imports
        >>> from pprint import pprint
        >>> import pandas as pd
        >>> from pyspark.sql import SparkSession, functions as F
        >>> from toolbox_pyspark.columns import get_columns
        >>>
        >>> # Instantiate Spark
        >>> spark = SparkSession.builder.getOrCreate()
        >>>
        >>> # Create data
        >>> df = (
        ...     spark
        ...     .createDataFrame(
        ...         pd.DataFrame(
        ...             {
        ...                 "a": (0, 1, 2, 3),
        ...                 "b": ["a", "b", "c", "d"],
        ...             }
        ...         )
        ...     )
        ...     .withColumns(
        ...         {
        ...             "c": F.lit("1").cast("int"),
        ...             "d": F.lit("2").cast("string"),
        ...             "e": F.lit("1.1").cast("float"),
        ...             "f": F.lit("1.2").cast("double"),
        ...             "g": F.lit("2022-01-01").cast("date"),
        ...             "h": F.lit("2022-02-01 01:00:00").cast("timestamp"),
        ...         }
        ...     )
        ... )
        >>>
        >>> # Check
        >>> df.show()
        >>> print(df.dtypes)
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+-----+-----+------------+---------------------+
        | a | b | c | d |   e |   f |          g |                   h |
        +---+---+---+---+-----+-----+------------+---------------------+
        | 0 | a | 1 | 2 | 1.1 | 1.2 | 2022-01-01 | 2022-02-01 01:00:00 |
        | 1 | b | 1 | 2 | 1.1 | 1.2 | 2022-01-01 | 2022-02-01 01:00:00 |
        | 2 | c | 1 | 2 | 1.1 | 1.2 | 2022-01-01 | 2022-02-01 01:00:00 |
        | 3 | d | 1 | 2 | 1.1 | 1.2 | 2022-01-01 | 2022-02-01 01:00:00 |
        +---+---+---+---+-----+-----+------------+---------------------+
        ```
        ```{.txt .text title="Terminal"}
        [
            ("a", "bigint"),
            ("b", "string"),
            ("c", "int"),
            ("d", "string"),
            ("e", "float"),
            ("f", "double"),
            ("g", "date"),
            ("h", "timestamp"),
        ]
        ```
        </div>

        ```{.py .python linenums="1" title="Example 1: Default params"}
        >>> print(get_columns(df).columns)
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        ["a", "b", "c", "d", "e", "f", "g", "h"]
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 2: Specific columns"}
        >>> print(get_columns(df, ["a", "b", "c"]).columns)
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        ["a", "b", "c"]
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 3: Single column as list"}
        >>> print(get_columns(df, ["a"]).columns)
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        ["a"]
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 4: Single column as string"}
        >>> print(get_columns(df, "a").columns)
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        ["a"]
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 5: All columns"}
        >>> print(get_columns(df, "all").columns)
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        ["a", "b", "c", "d", "e", "f", "g", "h"]
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 6: All str"}
        >>> print(get_columns(df, "all_str").columns)
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        ["b", "d"]
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 7: All int"}
        >>> print(get_columns(df, "all int").columns)
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        ["c"]
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 8: All float"}
        >>> print(get_columns(df, "all_decimal").columns)
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        ["e", "f"]
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 9: All numeric"}
        >>> print(get_columns(df, "all_numeric").columns)
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        ["c", "e", "f"]
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 10: All date"}
        >>> print(get_columns(df, "all_date").columns)
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        ["g"]
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 11: All datetime"}
        >>> print(get_columns(df, "all_datetime").columns)
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        ["h"]
        ```
        !!! success "Conclusion: Success."
        </div>
    """
    if columns is None:
        return dataframe.columns
    elif is_type(columns, str):
        if "all" in columns:
            if "str" in columns:
                return [
                    col for col, typ in dataframe.dtypes if typ in ("str", "string")
                ]
            elif "int" in columns:
                return [
                    col for col, typ in dataframe.dtypes if typ in ("int", "integer")
                ]
            elif "numeric" in columns:
                return [
                    col
                    for col, typ in dataframe.dtypes
                    if typ in ("int", "integer", "float", "double", "long")
                    or "decimal" in typ
                ]
            elif "float" in columns or "double" in columns or "decimal" in columns:
                return [
                    col
                    for col, typ in dataframe.dtypes
                    if typ in ("float", "double", "long") or "decimal" in typ
                ]
            elif "datetime" in columns or "timestamp" in columns:
                return [
                    col
                    for col, typ in dataframe.dtypes
                    if typ in ("datetime", "timestamp")
                ]
            elif "date" in columns:
                return [col for col, typ in dataframe.dtypes if typ in ["date"]]
            else:
                return dataframe.columns
        else:
            return [columns]
    else:
        return list(columns)


@typechecked
def get_columns_by_likeness(
    dataframe: psDataFrame,
    starts_with: Optional[str] = None,
    contains: Optional[str] = None,
    ends_with: Optional[str] = None,
    match_case: bool = False,
    operator: Literal["and", "or", "and not", "or not"] = "and",
) -> str_list:
    """
    !!! note "Summary"
        Extract the column names from a given `dataframe` based on text that the column name contains.

    ???+ abstract "Details"
        You can use any combination of `startswith`, `contains`, and `endswith`. Under the hood, these will be implemented with a number of internal `#!py lambda` functions to determine matches.

        The `operator` parameter determines how the conditions (`starts_with`, `contains`, `ends_with`) are combined:

        | Value | Description |
        |-------|-------------|
        | `"and"` | All conditions must be true.
        | `"or"` | At least one condition must be true.
        | `"and not"` | The first condition must be true and the second condition must be false.
        | `"or not"` | At least one condition must be true, but not all.

    Params:
        dataframe (psDataFrame):
            The `dataframe` from which to extract the column names.
        starts_with (Optional[str], optional):
            Extract any columns that starts with this `#!py str`.<br>
            Determined by using the `#!py str.startswith()` method.<br>
            Defaults to `#!py None`.
        contains (Optional[str], optional):
            Extract any columns that contains this `#!py str` anywhere within it.<br>
            Determined by using the `#!py in` keyword.<br>
            Defaults to `#!py None`.
        ends_with (Optional[str], optional):
            Extract any columns that ends with this `#!py str`.<br>
            Determined by using the `#!py str.endswith()` method.<br>
            Defaults to `#!py None`.
        match_case (bool, optional):
            If you want to ensure an exact match for the columns, set this to `#!py True`, else if you want to match the exact case for the columns, set this to `#!py False`.<br>
            Defaults to `#!py False`.
        operator (Literal["and", "or", "and not", "or not"], optional):
            The logical operator to place between the functions.<br>
            Only used when there are multiple values parsed to the parameters: `#!py starts_with`, `#!py contains`: `#!py ends_with`.<br>
            Defaults to `#!py and`.

    Returns:
        (str_list):
            The list of columns which match the criteria specified.

    ???+ example "Examples"

        ```{.py .python linenums="1" title="Set up"}
        >>> # Imports
        >>> import pandas as pd
        >>> from pyspark.sql import SparkSession
        >>> from toolbox_pyspark.columns import get_columns_by_likeness
        >>>
        >>> # Instantiate Spark
        >>> spark = SparkSession.builder.getOrCreate()
        >>>
        >>> # Create data
        >>> values = list(range(1, 6))
        >>> df = spark.createDataFrame(
        ...     pd.DataFrame(
        ...         {
        ...             "aaa": values,
        ...             "aab": values,
        ...             "aac": values,
        ...             "afa": values,
        ...             "afb": values,
        ...             "afc": values,
        ...             "bac": values,
        ...         }
        ...     )
        ... )
        >>>
        >>> # Check
        >>> df.show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +-----+-----+-----+-----+-----+-----+-----+
        | aaa | aab | aac | afa | afb | afc | bac |
        +-----+-----+-----+-----+-----+-----+-----+
        |   1 |   1 |   1 |   1 |   1 |   1 |   1 |
        |   2 |   2 |   2 |   2 |   2 |   2 |   2 |
        |   3 |   3 |   3 |   3 |   3 |   3 |   3 |
        |   4 |   4 |   4 |   4 |   4 |   4 |   4 |
        |   5 |   5 |   5 |   5 |   5 |   5 |   5 |
        +-----+-----+-----+-----+-----+-----+-----+
        ```
        </div>

        ```{.py .python linenums="1" title="Example 1: Starts With"}
        >>> print(get_columns_by_likeness(df, starts_with="a"))
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        ["aaa", "aab", "aac", "afa", "afb", "afc"]
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 2: Contains"}
        >>> print(get_columns_by_likeness(df, contains="f"))
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        ["afa", "afb", "afc"]
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 3: Ends With"}
        >>> print(get_columns_by_likeness(df, ends_with="c"))
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        ["aac", "afc", "bac"]
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 4: Starts With and Contains"}
        >>> print(get_columns_by_likeness(df, starts_with="a", contains="c"))
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        ["aac", "afc"]
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 5: Starts With and Ends With"}
        >>> print(get_columns_by_likeness(df, starts_with="a", ends_with="b"))
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        ["aab", "afb"]
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 6: Contains and Ends With"}
        >>> print(get_columns_by_likeness(df, contains="f", ends_with="b"))
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        ["afb"]
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 7: Starts With and Contains and Ends With"}
        >>> print(get_columns_by_likeness(df, starts_with="a", contains="f", ends_with="b"))
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        ["afb"]
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 8: Using 'or' Operator"}
        >>> print(get_columns_by_likeness(df, starts_with="a", operator="or", contains="f"))
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        ["aaa", "aab", "aac", "afa", "afb", "afc"]
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 9: Using 'and not' Operator"}
        >>> print(get_columns_by_likeness(df, starts_with="a", operator="and not", contains="f"))
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        ["aaa", "aab", "aac"]
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 10: Error Example 1"}
        >>> print(get_columns_by_likeness(df, starts_with=123))
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        TypeError: `starts_with` must be a `string` or `None`.
        ```
        !!! failure "Conclusion: Error."
        </div>

        ```{.py .python linenums="1" title="Example 11: Error Example 2"}
        >>> print(get_columns_by_likeness(df, operator="xor"))
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        ValueError: `operator` must be one of 'and', 'or', 'and not', 'or not'
        ```
        !!! failure "Conclusion: Error."
        </div>
    """

    # Columns
    cols: str_list = dataframe.columns
    if not match_case:
        cols = [col.upper() for col in cols]
        starts_with = starts_with.upper() if starts_with is not None else None
        contains = contains.upper() if contains is not None else None
        ends_with = ends_with.upper() if ends_with is not None else None

    # Parameters
    o_: Literal["and", "or", "and not", "or not"] = operator
    s_: bool = starts_with is not None
    c_: bool = contains is not None
    e_: bool = ends_with is not None

    # Functions
    _ops = {
        "and": lambda x, y: x and y,
        "or": lambda x, y: x or y,
        "and not": lambda x, y: x and not y,
        "or not": lambda x, y: x or not y,
    }
    _s = lambda col, s: col.startswith(s)
    _c = lambda col, c: c in col
    _e = lambda col, e: col.endswith(e)
    _sc = lambda col, s, c: _ops[o_](_s(col, s), _c(col, c))
    _se = lambda col, s, e: _ops[o_](_s(col, s), _e(col, e))
    _ce = lambda col, c, e: _ops[o_](_c(col, c), _e(col, e))
    _sce = lambda col, s, c, e: _ops[o_](_ops[o_](_s(col, s), _c(col, c)), _e(col, e))

    # Logic
    if s_ and not c_ and not e_:
        return [col for col in cols if _s(col, starts_with)]
    elif c_ and not s_ and not e_:
        return [col for col in cols if _c(col, contains)]
    elif e_ and not s_ and not c_:
        return [col for col in cols if _e(col, ends_with)]
    elif s_ and c_ and not e_:
        return [col for col in cols if _sc(col, starts_with, contains)]
    elif s_ and e_ and not c_:
        return [col for col in cols if _se(col, starts_with, ends_with)]
    elif c_ and e_ and not s_:
        return [col for col in cols if _ce(col, contains, ends_with)]
    elif s_ and c_ and e_:
        return [col for col in cols if _sce(col, starts_with, contains, ends_with)]
    else:
        return cols


# ---------------------------------------------------------------------------- #
#  Renaming                                                                 ####
# ---------------------------------------------------------------------------- #


@typechecked
def rename_columns(
    dataframe: psDataFrame,
    columns: Optional[Union[str, str_collection]] = None,
    string_function: str = "upper",
) -> psDataFrame:
    """
    !!! note "Summary"
        Use one of the common Python string functions to be applied to one or multiple columns.

    ???+ abstract "Details"
        The `string_function` must be a valid string method. For more info on available functions, see: https://docs.python.org/3/library/stdtypes.html#string-methods

    Params:
        dataframe (psDataFrame):
            The DataFrame to be updated.
        columns (Optional[Union[str, str_collection]], optional):
            The columns to be updated.<br>
            Must be a valid column on `dataframe`.<br>
            If not provided, will be applied to all columns.<br>
            It is also possible to parse the values `"all"`, which will also apply this function to all columns in `dataframe`.<br>
            Defaults to `None`.
        string_function (str, optional):
            The string function to be applied. Defaults to `"upper"`.

    Returns:
        (psDataFrame):
            The updated DataFrame.

    ???+ example "Examples"

        ```{.py .python linenums="1" title="Set up"}
        >>> # Import
        >>> import pandas as pd
        >>> from pyspark.sql import SparkSession
        >>> from toolbox_pyspark.columns import rename_columns
        >>>
        >>> # Instantiate Spark
        >>> spark = SparkSession.builder.getOrCreate()
        >>>
        >>> # Create data
        >>> df = spark.createDataFrame(
        ...     pd.DataFrame(
        ...         {
        ...             "a": [0, 1, 2, 3],
        ...             "b": ["a", "b", "c", "d"],
        ...             "c": ["c", "c", "c", "c"],
        ...             "d": ["d", "d", "d", "d"],
        ...         }
        ...     )
        ... )
        >>>
        >>> # Check
        >>> df.show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+
        | a | b | c | d |
        +---+---+---+---+
        | 0 | a | c | d |
        | 1 | b | c | d |
        | 2 | c | c | d |
        | 3 | d | c | d |
        +---+---+---+---+
        ```
        </div>

        ```{.py .python linenums="1" title="Example 1: Single column, default params"}
        >>> print(rename_columns(df, "a").columns)
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        ["A", "b", "c", "d"]
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 2: Single column, simple function"}
        >>> print(rename_columns(df, "a", "upper").columns)
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        ["A", "b", "c", "d"]
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 3: Single column, complex function"}
        >>> print(rename_columns(df, "a", "replace('b', 'test')").columns)
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        ["a", "test", "c", "d"]
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 4: Multiple columns"}
        >>> print(rename_columns(df, ["a", "b"]).columns)
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        ["A", "B", "c", "d"]
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 5: Default function over all columns"}
        >>> print(rename_columns(df).columns)
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        ["A", "B", "C", "D"]
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 6: Complex function over multiple columns"}
        >>> print(rename_columns(df, ["a", "b"], "replace('b', 'test')").columns)
        ```
        <div class="result" markdown>
        ```{.sh .shell title="Terminal"}
        ["a", "test", "c", "d"]
        ```
        !!! success "Conclusion: Success."
        </div>

    ??? tip "See Also"
        - [`assert_columns_exists()`][toolbox_pyspark.checks.assert_columns_exists]
        - [`assert_column_exists()`][toolbox_pyspark.checks.assert_column_exists]
    """
    columns = get_columns(dataframe, columns)
    assert_columns_exists(dataframe=dataframe, columns=columns, match_case=True)
    cols_exprs: dict[str, str] = {
        col: eval(
            f"'{col}'.{string_function}{'()' if not string_function.endswith(')') else ''}"
        )
        for col in columns
    }
    return dataframe.withColumnsRenamed(cols_exprs)


# ---------------------------------------------------------------------------- #
#  Reordering                                                               ####
# ---------------------------------------------------------------------------- #


@typechecked
def reorder_columns(
    dataframe: psDataFrame,
    new_order: Optional[str_collection] = None,
    missing_columns_last: bool = True,
    key_columns_position: Optional[Literal["first", "last"]] = "first",
) -> psDataFrame:
    """
    !!! note "Summary"
        Reorder the columns in a given DataFrame in to a custom order, or to put the `key_` columns at the end (that is, to the far right) of the dataframe.

    ???+ abstract "Details"
        The decision flow chart is as follows:

        ```mermaid
        graph TD
            a([begin])
            z([end])
            b{{new_order}}
            c{{missing_cols_last}}
            d{{key_cols_position}}
            g[cols = dataframe.columns]
            h[cols = new_order]
            i[cols += missing_cols]
            j[cols = non_key_cols + key_cols]
            k[cols = key_cols + non_key_cols]
            l["return dataframe.select(cols)"]
            a --> b
            b --is not None--> h --> c
            b --is None--> g --> d
            c --False--> l
            c --True--> i ----> l
            d --"first"--> k ---> l
            d --"last"---> j --> l
            d --None--> l
            l --> z
        ```

    Params:
        dataframe (psDataFrame):
            The DataFrame to update
        new_order (Optional[Union[str, str_list, str_tuple, str_set]], optional):
            The custom order for the columns on the order.<br>
            Defaults to `#!py None`.
        missing_columns_last (bool, optional):
            For any columns existing on `#!py dataframes.columns`, but missing from `#!py new_order`, if `#!py missing_columns_last=True`, then include those missing columns to the right of the dataframe, in the same order that they originally appear.<br>
            Defaults to `#!py True`.
        key_columns_position (Optional[Literal["first", "last"]], optional):
            Where should the `#!py "key_*"` columns be located?.<br>

            - If `#!py "first"`, then they will be relocated to the start of the dataframe, before all other columns.
            - If `#!py "last"`, then they will be relocated to the end of the dataframe, after all other columns.
            - If `#!py None`, they they will remain their original order.

            Regardless of their position, their original order will be maintained.
            Defaults to `#!py "first"`.

    Raises:
        TypeError:
            If any of the inputs parsed to the parameters of this function are not the correct type. Uses the [`@typeguard.typechecked`](https://typeguard.readthedocs.io/en/stable/api.html#typeguard.typechecked) decorator.

    Returns:
        (psDataFrame):
            The updated DataFrame.

    ???+ example "Examples"

        ```{.py .python linenums="1" title="Set up"}
        >>> # Imports
        >>> import pandas as pd
        >>> from pyspark.sql import SparkSession
        >>> from toolbox_pyspark.columns import reorder_columns
        >>>
        >>> # Instantiate Spark
        >>> spark = SparkSession.builder.getOrCreate()
        >>>
        >>> # Create data
        >>> df = spark.createDataFrame(
        ...     pd.DataFrame(
        ...         {
        ...             "a": [0, 1, 2, 3],
        ...             "b": ["a", "b", "c", "d"],
        ...             "key_a": ["0", "1", "2", "3"],
        ...             "c": ["1", "1", "1", "1"],
        ...             "d": ["2", "2", "2", "2"],
        ...             "key_c": ["1", "1", "1", "1"],
        ...             "key_e": ["3", "3", "3", "3"],
        ...         }
        ...     )
        ... )
        >>>
        >>> # Check
        >>> df.show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+-------+---+---+-------+-------+
        | a | b | key_a | c | d | key_c | key_e |
        +---+---+-------+---+---+-------+-------+
        | 0 | a |     0 | 1 | 2 |     1 |     3 |
        | 1 | b |     1 | 1 | 2 |     1 |     3 |
        | 2 | c |     2 | 1 | 2 |     1 |     3 |
        | 3 | d |     3 | 1 | 2 |     1 |     3 |
        +---+---+-------+---+---+-------+-------+
        ```
        </div>

        ```{.py .python linenums="1" title="Example 1: Default config"}
        >>> new_df = reorder_columns(dataframe=df)
        >>> new_df.show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +-------+-------+-------+---+---+---+---+
        | key_a | key_c | key_e | a | b | c | d |
        +-------+-------+-------+---+---+---+---+
        |     0 |     1 |     3 | 0 | a | 1 | 2 |
        |     1 |     1 |     3 | 1 | b | 1 | 2 |
        |     2 |     1 |     3 | 2 | c | 1 | 2 |
        |     3 |     1 |     3 | 3 | d | 1 | 2 |
        +-------+-------+-------+---+---+---+---+
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 2: Custom order"}
        >>> new_df = reorder_columns(
        ...     dataframe=df,
        ...     new_order=["key_a", "key_c", "b", "key_e", "a", "c", "d"],
        ... )
        >>> new_df.show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +-------+-------+---+-------+---+---+---+
        | key_a | key_c | b | key_e | a | c | d |
        +-------+-------+---+-------+---+---+---+
        |     0 |     1 | a |     3 | 0 | 1 | 2 |
        |     1 |     1 | b |     3 | 1 | 1 | 2 |
        |     2 |     1 | c |     3 | 2 | 1 | 2 |
        |     3 |     1 | d |     3 | 3 | 1 | 2 |
        +-------+-------+---+-------+---+---+---+
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 3: Custom order, include missing columns"}
        >>> new_df = reorder_columns(
        ...     dataframe=df,
        ...     new_order=["key_a", "key_c", "a", "b"],
        ...     missing_columns_last=True,
        ...     )
        >>> new_df.show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +-------+-------+---+---+-------+---+---+
        | key_a | key_c | a | b | key_e | c | d |
        +-------+-------+---+---+-------+---+---+
        |     0 |     1 | 0 | a |     3 | 1 | 2 |
        |     1 |     1 | 1 | b |     3 | 1 | 2 |
        |     2 |     1 | 2 | c |     3 | 1 | 2 |
        |     3 |     1 | 3 | d |     3 | 1 | 2 |
        +-------+-------+---+---+-------+---+---+
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 4: Custom order, exclude missing columns"}
        >>> new_df = reorder_columns(
        ...     dataframe=df,
        ...     new_order=["key_a", "key_c", "a", "b"],
        ...     missing_columns_last=False,
        ... )
        >>> new_df.show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +-------+-------+---+---+
        | key_a | key_c | a | b |
        +-------+-------+---+---+
        |     0 |     1 | 0 | a |
        |     1 |     1 | 1 | b |
        |     2 |     1 | 2 | c |
        |     3 |     1 | 3 | d |
        +-------+-------+---+---+
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 5: Keys last"}
        >>> new_df = reorder_columns(
        ...     dataframe=df,
        ...     key_columns_position="last",
        ... )
        >>> new_df.show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+-------+-------+-------+
        | a | b | c | d | key_a | key_c | key_e |
        +---+---+---+---+-------+-------+-------+
        | 0 | a | 1 | 2 |     0 |     1 |     3 |
        | 1 | b | 1 | 2 |     1 |     1 |     3 |
        | 2 | c | 1 | 2 |     2 |     1 |     3 |
        | 3 | d | 1 | 2 |     3 |     1 |     3 |
        +---+---+---+---+-------+-------+-------+
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 6: Keys first"}
        >>> new_df = reorder_columns(
        ...     dataframe=df,
        ...     key_columns_position="first",
        ... )
        >>> new_df.show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +-------+-------+-------+---+---+---+---+
        | key_a | key_c | key_e | a | b | c | d |
        +-------+-------+-------+---+---+---+---+
        |     0 |     1 |     3 | 0 | a | 1 | 2 |
        |     1 |     1 |     3 | 1 | b | 1 | 2 |
        |     2 |     1 |     3 | 2 | c | 1 | 2 |
        |     3 |     1 |     3 | 3 | d | 1 | 2 |
        +-------+-------+-------+---+---+---+---+
        ```
        !!! success "Conclusion: Success."
        </div>
    """
    df_cols: str_list = dataframe.columns
    if new_order is not None:
        cols: str_list = get_columns(dataframe, new_order)
        if missing_columns_last:
            cols += [col for col in df_cols if col not in new_order]
    else:
        non_key_cols: str_list = [
            col for col in df_cols if not col.lower().startswith("key_")
        ]
        key_cols: str_list = [col for col in df_cols if col.lower().startswith("key_")]
        if key_columns_position == "first":
            cols = key_cols + non_key_cols
        elif key_columns_position == "last":
            cols = non_key_cols + key_cols
        else:
            cols = df_cols
    return dataframe.select(cols)


# ---------------------------------------------------------------------------- #
#  Deleting                                                                 ####
# ---------------------------------------------------------------------------- #


@typechecked
def delete_columns(
    dataframe: psDataFrame,
    columns: Union[str, str_collection],
    missing_column_handler: Literal["raise", "warn", "pass"] = "pass",
) -> psDataFrame:
    """
    !!! note "Summary"
        For a given `#!py dataframe`, delete the columns listed in `columns`.

    ???+ abstract "Details"
        You can use `#!py missing_columns_handler` to specify how to handle missing columns.

    Params:
        dataframe (psDataFrame):
            The dataframe from which to delete the columns
        columns (Union[str, str_collection]):
            The list of columns to delete.
        missing_column_handler (Literal["raise", "warn", "pass"], optional):
            How to handle any columns which are missing from `#!py dataframe.columns`.

            If _any_ columns in `columns` are missing from `#!py dataframe.columns`, then the following will happen for each option:

            | Option | Result |
            |--------|--------|
            | `#!py "raise"` | An `#!py ColumnDoesNotExistError` exception will be raised
            | `#!py "warn"` | An `#!py ColumnDoesNotExistWarning` warning will be raised
            | `#!py "pass"` | Nothing will be raised

            Defaults to `#!py "pass"`.

    Raises:
        TypeError:
            If any of the inputs parsed to the parameters of this function are not the correct type. Uses the [`@typeguard.typechecked`](https://typeguard.readthedocs.io/en/stable/api.html#typeguard.typechecked) decorator.
        ColumnDoesNotExistError:
            If any of the `#!py columns` do not exist within `#!py dataframe.columns`.

    Returns:
        (psDataFrame):
            The updated `#!py dataframe`, with the columns listed in `#!py columns` having been removed.

    ???+ example "Examples"

        ```{.py .python linenums="1" title="Set up"}
        >>> # Imports
        >>> import pandas as pd
        >>> from pyspark.sql import SparkSession
        >>> from toolbox_pyspark.columns import delete_columns
        >>>
        >>> # Instantiate Spark
        >>> spark = SparkSession.builder.getOrCreate()
        >>>
        >>> # Create data
        >>> df = spark.createDataFrame(
        ...     pd.DataFrame(
        ...         {
        ...             "a": [0, 1, 2, 3],
        ...             "b": ["a", "b", "c", "d"],
        ...             "c": ["c", "c", "c", "c"],
        ...             "d": ["d", "d", "d", "d"],
        ...         }
        ...     )
        ... )
        >>>
        >>> # Check
        >>> df.show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+
        | a | b | c | d |
        +---+---+---+---+
        | 0 | a | c | d |
        | 1 | b | c | d |
        | 2 | c | c | d |
        | 3 | d | c | d |
        +---+---+---+---+
        ```
        </div>

        ```{.py .python linenums="1" title="Example 1: Single column"}
        >>> df.transform(delete_columns, "a").show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+
        | b | c | d |
        +---+---+---+
        | a | c | d |
        | b | c | d |
        | c | c | d |
        | d | c | d |
        +---+---+---+
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 2: Multiple columns"}
        >>> df.transform(delete_columns, ["a", "b"]).show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+
        | c | d |
        +---+---+
        | c | d |
        | c | d |
        | c | d |
        | c | d |
        +---+---+
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 3: Single column missing, raises error"}
        >>> (
        ...     df.transform(
        ...         delete_columns,
        ...         columns="z",
        ...         missing_column_handler="raise",
        ...     )
        ...     .show()
        ... )
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        ColumnDoesNotExistError: Columns ["z"] do not exist in "dataframe".
        Try one of: ["a", "b", "c", "d"]
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 4: Multiple columns, one missing, raises error"}
        >>> (
        ...     df.transform(
        ...         delete_columns,
        ...         columns=["a", "b", "z"],
        ...         missing_column_handler="raise",
        ...     )
        ...     .show()
        ... )
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        ColumnDoesNotExistError: Columns ["z"] do not exist in "dataframe".
        Try one of: ["a", "b", "c", "d"]
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 5: Multiple columns, all missing, raises error"}
        >>> (
        ...     df.transform(
        ...         delete_columns,
        ...         columns=["x", "y", "z"],
        ...         missing_column_handler="raise",
        ...     )
        ...     .show()
        ... )
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        ColumnDoesNotExistError: Columns ["x", "y", "z"] do not exist in "dataframe".
        Try one of: ["a", "b", "c", "d"]
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 6: Single column missing, raises warning"}
        >>> (
        ...     df.transform(
        ...         delete_columns,
        ...         columns="z",
        ...         missing_column_handler="warn",
        ...     )
        ...     .show()
        ... )
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        ColumnDoesNotExistWarning: Columns missing from "dataframe": ["z"].
        Will still proceed to delete columns that do exist.
        ```
        ```{.txt .text title="Terminal"}
        +---+---+---+---+
        | a | b | c | d |
        +---+---+---+---+
        | 0 | a | c | d |
        | 1 | b | c | d |
        | 2 | c | c | d |
        | 3 | d | c | d |
        +---+---+---+---+
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 7: Multiple columns, one missing, raises warning"}
        >>> (
        ...     df.transform(
        ...         delete_columns,
        ...         columns=["a", "b", "z"],
        ...         missing_column_handler="warn",
        ...     )
        ...     .show()
        ... )
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        ColumnDoesNotExistWarning: Columns missing from "dataframe": ["z"].
        Will still proceed to delete columns that do exist.
        ```
        ```{.txt .text title="Terminal"}
        +---+---+
        | c | d |
        +---+---+
        | c | d |
        | c | d |
        | c | d |
        | c | d |
        +---+---+
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 8: Multiple columns, all missing, raises warning"}
        >>> (
        ...     df.transform(
        ...         delete_columns,
        ...         columns=["x", "y", "z"],
        ...         missing_column_handler="warn",
        ...     )
        ...     .show()
        ... )
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        ColumnDoesNotExistWarning: Columns missing from "dataframe": ["x", "y", "z"].
        Will still proceed to delete columns that do exist.
        ```
        ```{.txt .text title="Terminal"}
        +---+---+---+---+
        | a | b | c | d |
        +---+---+---+---+
        | 0 | a | c | d |
        | 1 | b | c | d |
        | 2 | c | c | d |
        | 3 | d | c | d |
        +---+---+---+---+
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 9: Single column missing, nothing raised"}
        >>> (
        ...     df.transform(
        ...         delete_columns,
        ...         columns="z",
        ...         missing_column_handler="pass",
        ...     )
        ...     .show()
        ... )
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+
        | a | b | c | d |
        +---+---+---+---+
        | 0 | a | c | d |
        | 1 | b | c | d |
        | 2 | c | c | d |
        | 3 | d | c | d |
        +---+---+---+---+
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 10: Multiple columns, one missing, nothing raised"}
        >>> (
        ...     df.transform(
        ...         delete_columns,
        ...         columns=["a", "b", "z"],
        ...         missing_column_handler="pass",
        ...     )
        ...     .show()
        ... )
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+
        | c | d |
        +---+---+
        | c | d |
        | c | d |
        | c | d |
        | c | d |
        +---+---+
        ```
        !!! success "Conclusion: Success."
        </div>

        ```{.py .python linenums="1" title="Example 11: Multiple columns, all missing, nothing raised"}
        >>> (
        ...     df.transform(
        ...         delete_columns,
        ...         columns=["x", "y", "z"],
        ...         missing_column_handler="pass",
        ...     )
        ...     .show()
        ... )
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+
        | a | b | c | d |
        +---+---+---+---+
        | 0 | a | c | d |
        | 1 | b | c | d |
        | 2 | c | c | d |
        | 3 | d | c | d |
        +---+---+---+---+
        ```
        !!! success "Conclusion: Success."
        </div>
    """
    columns = get_columns(dataframe, columns)
    if missing_column_handler == "raise":
        assert_columns_exists(dataframe=dataframe, columns=columns)
    elif missing_column_handler == "warn":
        warn_columns_missing(dataframe=dataframe, columns=columns)
    elif missing_column_handler == "pass":
        pass
    return dataframe.select([col for col in dataframe.columns if col not in columns])



# ============================================================================ #
#                                                                              #
#     Title   : Dimensions                                                     #
#     Purpose : Check the dimensions of a`pyspark` `dataframes`.               #
#                                                                              #
# ============================================================================ #


# ---------------------------------------------------------------------------- #
#                                                                              #
#     Overview                                                              ####
#                                                                              #
# ---------------------------------------------------------------------------- #


# ---------------------------------------------------------------------------- #
#  Description                                                              ####
# ---------------------------------------------------------------------------- #


"""
!!! note "Summary"
    The `dimensions` module is used for checking the dimensions of `pyspark` `dataframe`'s.
"""


# ---------------------------------------------------------------------------- #
#                                                                              #
#     Setup                                                                 ####
#                                                                              #
# ---------------------------------------------------------------------------- #


# ---------------------------------------------------------------------------- #
#  Imports                                                                  ####
# ---------------------------------------------------------------------------- #


# ## Python StdLib Imports ----
from typing import Dict, Optional, Union

# ## Python Third Party Imports ----
import numpy as np
from pandas import DataFrame as pdDataFrame
from pyspark.sql import DataFrame as psDataFrame
from toolbox_python.collection_types import str_list
from typeguard import typechecked


# ---------------------------------------------------------------------------- #
#  Exports                                                                  ####
# ---------------------------------------------------------------------------- #


__all__: str_list = ["get_dims", "get_dims_of_tables"]


# ---------------------------------------------------------------------------- #
#                                                                              #
#     Functions                                                             ####
#                                                                              #
# ---------------------------------------------------------------------------- #


# ---------------------------------------------------------------------------- #
#  Functions                                                                ####
# ---------------------------------------------------------------------------- #


@typechecked
def get_dims(
    dataframe: psDataFrame,
    use_names: bool = True,
    use_comma: bool = True,
) -> Union[dict[str, str], dict[str, int], tuple[str, str], tuple[int, int]]:
    """
    !!! note "Summary"
        Extract the dimensions of a given `dataframe`.

    Params:
        dataframe (psDataFrame):
            The table to check.
        use_names (bool, optional):
            Whether or not to add `names` to the returned object.<br>
            If `#!py True`, then will return a `#!py dict` with two keys only, for the number of `rows` and `cols`.<br>
            Defaults to `#!py True`.
        use_comma (bool, optional):
            Whether or not to add a comma `,` to the returned object.<br>
            Defaults to `#!py True`.

    Raises:
        TypeError:
            If any of the inputs parsed to the parameters of this function are not the correct type. Uses the [`@typeguard.typechecked`](https://typeguard.readthedocs.io/en/stable/api.html#typeguard.typechecked) decorator.

    Returns:
        (Union[Dict[str, Union[str, int]], tuple[str, ...], tTuple[int, ...]]):
            The dimensions of the given `dataframe`.

    ???+ example "Examples"

        ```{.py .python linenums="1" title="Set up"}
        >>> import pandas as pd
        >>> from pyspark.sql import SparkSession
        >>> from toolbox_pyspark.dimensions import get_dims
        >>> spark = SparkSession.builder.getOrCreate()
        >>> df = spark.createDataFrame(
        ...     pd.DataFrame({
        ...         'a': range(5000),
        ...         'b': range(5000),
        ...     })
        ... )
        ```

        ```{.py .python linenums="1" title="Check"}
        >>> print(df.count())
        >>> print(len(df.columns))
        ```
        <div class="result" markdown>
        ```{.txt .text}
        5000
        ```

        ```{.txt .text}
        2
        ```
        </div>

        ```{.py .python linenums="1" title="Names and commas"}
        >>> print(get_dims(dataframe=df, use_names=True, use_commas=True))
        ```
        <div class="result" markdown>
        ```{.py .python}
        {"rows": "5,000", "cols": "2"}
        ```
        </div>

        ```{.py .python linenums="1" title="Names but no commas"}
        >>> print(get_dims(dataframe=df, use_names=True, use_commas=False))
        ```
        <div class="result" markdown>
        ```{.py .python}
        {"rows": 5000, "cols": 2}
        ```
        </div>

        ```{.py .python linenums="1" title="Commas but no names"}
        >>> print(get_dims(dataframe=df, use_names=False, use_commas=True))
        ```
        <div class="result" markdown>
        ```{.py .python}
        ("5,000", "2")
        ```
        </div>

        ```{.py .python linenums="1" title="Neither names nor commas"}
        >>> print(get_dims(dataframe=df, use_names=False, use_commas=False))
        ```
        <div class="result" markdown>
        ```{.py .python}
        (5000, 2)
        ```
        </div>
    """
    dims: tuple[int, int] = (dataframe.count(), len(dataframe.columns))
    if use_names and use_comma:
        return {"rows": f"{dims[0]:,}", "cols": f"{dims[1]:,}"}
    elif use_names and not use_comma:
        return {"rows": dims[0], "cols": dims[1]}
    elif not use_names and use_comma:
        return (f"{dims[0]:,}", f"{dims[1]:,}")
    else:
        return dims


@typechecked
def get_dims_of_tables(
    tables: str_list,
    scope: Optional[dict] = None,
    use_comma: bool = True,
) -> pdDataFrame:
    """
    !!! note "Summary"
        Take in a list of the names of some tables, and for each of them, check their dimensions.

    ???+ abstract "Details"
        This function will check against the `#!py global()` scope. So you need to be careful if you're dealing with massive amounts of data in memory.

    Params:
        tables (List[str]):
            The list of the tables that will be checked.
        scope (dict, optional):
            This is the scope against which the tables will be checked.<br>
            If `#!py None`, then it will use the `#!py global()` scope by default..<br>
            Defaults to `#!py None`.
        use_comma (bool, optional):
            Whether or not the dimensions from the tables should be formatted as a string with a comma as the thousandths delimiter.<br>
            Defaults to `#!py True`.

    Raises:
        TypeError:
            If any of the inputs parsed to the parameters of this function are not the correct type. Uses the [`@typeguard.typechecked`](https://typeguard.readthedocs.io/en/stable/api.html#typeguard.typechecked) decorator.

    Returns:
        (pdDataFrame):
            A `pandas` `dataframe` with four columns: `#!py ["table", "type", "rows", "cols"]`.

    ???+ example "Examples"

        ```{.py .python linenums="1" title="Set up"}
        >>> import pandas as pd
        >>> from pyspark.sql import SparkSession
        >>> from toolbox_pyspark.dimensions import get_dims_of_tables, get_dims
        >>> spark = SparkSession.builder.getOrCreate()
        >>> df1 = spark.createDataFrame(
        ...     pd.DataFrame({
        ...         'a': range(5000),
        ...         'b': range(5000),
        ...     })
        ... )
        >>> df2 = spark.createDataFrame(
        ...     pd.DataFrame({
        ...         'a': range(10000),
        ...         'b': range(10000),
        ...         'c': range(10000),
        ...     })
        ... )
        ```

        ```{.py .python linenums="1" title="Check"}
        >>> print(get_dims(df1))
        >>> print(get_dims(df1))
        ```
        <div class="result" markdown>
        ```{.txt .text}
        {"rows": "5000", "cols": "2"}
        ```

        ```{.txt .text}
        {"rows": "10000", "cols": "3"}
        ```
        </div>

        ```{.py .python linenums="1" title="Basic usage"}
        >>> print(get_dims_of_tables(['df1', 'df2']))
        ```
        <div class="result" markdown>
        ```{.txt .text}
          table type  rows cols
        0   df1      5,000    2
        1   df2      1,000    3
        ```
        </div>

        ```{.py .python linenums="1" title="No commas"}
        >>> print(get_dims_of_tables(['df1', 'df2'], use_commas=False))
        ```
        <div class="result" markdown>
        ```{.txt .text}
          table type rows cols
        0   df1      5000    2
        1   df2      1000    3
        ```
        </div>

        ```{.py .python linenums="1" title="Missing DF"}
        >>> display(get_dims_of_tables(['df1', 'df2', 'df3'], use_comma=False))
        ```
        <div class="result" markdown>
        ```{.txt .text}
          table type rows cols
        0   df1      5000    2
        1   df2      1000    3
        1   df3       NaN  NaN
        ```
        </div>

    ??? info "Notes"
        - The first column of the returned table is the name of the table from the `scope` provided.
        - The second column of the returned table is the `type` of the table. That is, whether the table is one of `#!py ["prd", "arc", "acm"]`, which are for 'production', 'archive', accumulation' categories. This is designated by the table containing an underscore (`_`), and having a suffic of either one of: `#!py "prd"`, `#!py "arc"`, or `#!py "acm"`. If the table does not contain this info, then the value in this second column will just be blank.
        - If one of the tables given in the `tables` list does not exist in the `scope`, then the values given in the `rows` and `cols` columns will either be the values: `#!py np.nan` or `#!py "Did not load"`.
    """
    sizes: Dict[str, list] = {
        "table": list(),
        "type": list(),
        "rows": list(),
        "cols": list(),
    }
    rows: Union[str, int, float]
    cols: Union[str, int, float]
    for tbl, typ in [
        (
            table.rsplit("_", 1)
            if "_" in table and table.endswith(("acm", "arc", "prd"))
            else (table, "")
        )
        for table in tables
    ]:
        try:
            tmp: psDataFrame = eval(
                f"{tbl}{f'_{typ}' if typ!='' else ''}",
                globals() if scope is None else scope,
            )
            rows, cols = get_dims(tmp, use_names=False, use_comma=use_comma)
        except Exception:
            if use_comma:
                rows = cols = "Did not load"
            else:
                rows = cols = np.nan
        sizes["table"].append(tbl)
        sizes["type"].append(typ)
        sizes["rows"].append(rows)
        sizes["cols"].append(cols)
    return pdDataFrame(sizes)



# ============================================================================ #
#                                                                              #
#     Title   : Dataframe Cleaning                                             #
#     Purpose : Clean, fix, and fetch various aspects on a given DataFrame.    #
#                                                                              #
# ============================================================================ #


# ---------------------------------------------------------------------------- #
#                                                                              #
#     Overview                                                              ####
#                                                                              #
# ---------------------------------------------------------------------------- #


# ---------------------------------------------------------------------------- #
#  Description                                                              ####
# ---------------------------------------------------------------------------- #


"""
!!! note "Summary"
    The `cleaning` module is used to clean, fix, and fetch various aspects on a given DataFrame.
"""


# ---------------------------------------------------------------------------- #
#                                                                              #
#     Setup                                                                 ####
#                                                                              #
# ---------------------------------------------------------------------------- #


# ---------------------------------------------------------------------------- #
#  Imports                                                                  ####
# ---------------------------------------------------------------------------- #


# ## Python StdLib Imports ----
from typing import Optional, Union

# ## Python Third Party Imports ----
from numpy import ndarray as npArray
from pandas import DataFrame as pdDataFrame
from pyspark.sql import (
    Column,
    DataFrame as psDataFrame,
    SparkSession,
    functions as F,
    types as T,
)
from toolbox_python.checkers import is_type
from toolbox_python.collection_types import str_collection, str_list
from toolbox_python.lists import flatten
from typeguard import typechecked

# ## Local First Party Imports ----
from toolbox_pyspark.checks import assert_column_exists, assert_columns_exists
from toolbox_pyspark.columns import get_columns
from toolbox_pyspark.constants import (
    LITERAL_LIST_OBJECT_NAMES,
    LITERAL_NUMPY_ARRAY_NAMES,
    LITERAL_PANDAS_DATAFRAME_NAMES,
    LITERAL_PYSPARK_DATAFRAME_NAMES,
    VALID_LIST_OBJECT_NAMES,
    VALID_NUMPY_ARRAY_NAMES,
    VALID_PANDAS_DATAFRAME_NAMES,
    VALID_PYAPARK_JOIN_TYPES,
    VALID_PYSPARK_DATAFRAME_NAMES,
    WHITESPACE_CHARACTERS as WHITESPACES,
)


# ---------------------------------------------------------------------------- #
#  Exports                                                                  ####
# ---------------------------------------------------------------------------- #


__all__: str_list = [
    "create_empty_dataframe",
    "keep_first_record_by_columns",
    "convert_dataframe",
    "get_column_values",
    "update_nullability",
    "trim_spaces_from_column",
    "trim_spaces_from_columns",
    "apply_function_to_column",
    "apply_function_to_columns",
    "drop_matching_rows",
]


# ---------------------------------------------------------------------------- #
#                                                                              #
#     Functions                                                             ####
#                                                                              #
# ---------------------------------------------------------------------------- #


# ---------------------------------------------------------------------------- #
#  Empty DataFrame                                                          ####
# ---------------------------------------------------------------------------- #


@typechecked
def create_empty_dataframe(spark_session: SparkSession) -> psDataFrame:
    return spark_session.createDataFrame([], T.StructType([]))


# ---------------------------------------------------------------------------- #
#  Column processes                                                         ####
# ---------------------------------------------------------------------------- #


@typechecked
def keep_first_record_by_columns(
    dataframe: psDataFrame,
    columns: Union[str, str_collection],
) -> psDataFrame:
    """
    !!! note "Summary"
        For a given Spark `#!py DataFrame`, keep the first record given by the column(s) specified in `#!py columns`.

    ???+ abstract "Details"
        The necessity for this function arose when we needed to perform a `#!py distinct()` function for a given `#!py DataFrame`; however, we still wanted to retain data provided in the other columns.

    Params:
        dataframe (psDataFrame):
            The DataFrame that you want to filter.
        columns (Union[str, str_collection]):
            The single or multiple columns by which you want to extract the distinct values from.

    Raises:
        TypeError:
            If any of the inputs parsed to the parameters of this function are not the correct type. Uses the [`@typeguard.typechecked`](https://typeguard.readthedocs.io/en/stable/api.html#typeguard.typechecked) decorator.
        ColumnDoesNotExistError:
            If any of the `#!py columns` do not exist within `#!py dataframe.columns`.

    Returns:
        (psDataFrame):
            The updated dataframe, retaining only the first unique set of records from the columns specified in `#!py columns`.

    ???+ example "Examples"

        ```{.py .python linenums="1" title="Set up"}
        >>> # Imports
        >>> import pandas as pd
        >>> from pyspark.sql import SparkSession
        >>> from toolbox_pyspark.cleaning import keep_first_record_by_columns
        >>>
        >>> # Instantiate Spark
        >>> spark = SparkSession.builder.getOrCreate()
        >>>
        >>> # Create data
        >>> df = spark.createDataFrame(
        ...     pd.DataFrame(
        ...         {
        ...             "a": [1, 2, 3, 4],
        ...             "b": ["a", "b", "c", "d"],
        ...             "c": [1, 1, 2, 2],
        ...             "d": [1, 2, 2, 2],
        ...             "e": [1, 1, 2, 3],
        ...         }
        ...     )
        ... )
        >>>
        >>> # Check
        >>> df.show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+---+
        | a | b | c | d | e |
        +---+---+---+---+---+
        | 1 | a | 1 | 1 | 1 |
        | 2 | b | 1 | 2 | 1 |
        | 3 | c | 2 | 2 | 2 |
        | 4 | d | 2 | 2 | 3 |
        +---+---+---+---+---+
        ```
        </div>

        ```{.py .python linenums="1" title="Example 1: Distinct by the `c` column"}
        >>> keep_first_record_by_columns(df, "c").show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+---+
        | a | b | c | d | e |
        +---+---+---+---+---+
        | 1 | a | 1 | 1 | 1 |
        | 3 | c | 2 | 2 | 2 |
        +---+---+---+---+---+
        ```
        !!! success "Conclusion: Successfully kept first records by the `c` column."
        </div>

        ```{.py .python linenums="1" title="Example 2: Distinct by the `d` column"}
        >>> keep_first_record_by_columns(df, "d").show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+---+
        | a | b | c | d | e |
        +---+---+---+---+---+
        | 1 | a | 1 | 1 | 1 |
        | 2 | b | 1 | 2 | 1 |
        +---+---+---+---+---+
        ```
        !!! success "Conclusion: Successfully kept first records by the `d` column."
        </div>

        ```{.py .python linenums="1" title="Example 3: Distinct by the `e` column"}
        >>> keep_first_record_by_columns(df, "e").show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+---+
        | a | b | c | d | e |
        +---+---+---+---+---+
        | 1 | a | 1 | 1 | 1 |
        | 3 | c | 2 | 2 | 2 |
        | 4 | d | 2 | 2 | 3 |
        +---+---+---+---+---+
        ```
        !!! success "Conclusion: Successfully kept first records by the `e` column."
        </div>

        ```{.py .python linenums="1" title="Example 4: Distinct by the `c` & `d` columns"}
        >>> keep_first_record_by_columns(df, ["c", "d"]).show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+---+
        | a | b | c | d | e |
        +---+---+---+---+---+
        | 1 | a | 1 | 1 | 1 |
        | 2 | b | 1 | 2 | 1 |
        | 3 | c | 2 | 2 | 2 |
        +---+---+---+---+---+
        ```
        !!! success "Conclusion: Successfully kept first records by the `c` & `d` columns."
        </div>

        ```{.py .python linenums="1" title="Example 5: Distinct by the `c` & `e` columns"}
        >>> keep_first_record_by_columns(df, ["c", "e"]).show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+---+
        | a | b | c | d | e |
        +---+---+---+---+---+
        | 1 | a | 1 | 1 | 1 |
        | 3 | c | 2 | 2 | 2 |
        | 4 | d | 2 | 2 | 3 |
        +---+---+---+---+---+
        ```
        !!! success "Conclusion: Successfully kept first records by the `c` & `e` columns."
        </div>

        ```{.py .python linenums="1" title="Example 6: Distinct by the `d` & `e` columns"}
        >>> keep_first_record_by_columns(df, ["d", "e"]).show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+---+
        | a | b | c | d | e |
        +---+---+---+---+---+
        | 1 | a | 1 | 1 | 1 |
        | 2 | b | 1 | 2 | 1 |
        | 3 | c | 2 | 2 | 2 |
        | 4 | d | 2 | 2 | 3 |
        +---+---+---+---+---+
        ```
        !!! success "Conclusion: Successfully kept first records by the `d` & `e` columns."
        </div>

        ```{.py .python linenums="1" title="Example 7: Distinct by the `c`, `d` & `e` columns"}
        >>> keep_first_record_by_columns(df, ["c", "d", "e"]).show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+---+
        | a | b | c | d | e |
        +---+---+---+---+---+
        | 1 | a | 1 | 1 | 1 |
        | 2 | b | 1 | 2 | 1 |
        | 3 | c | 2 | 2 | 2 |
        | 4 | d | 2 | 2 | 3 |
        +---+---+---+---+---+
        ```
        !!! success "Conclusion: Successfully kept first records by the `c`, `d` & `e` columns."
        !!! failure "Conclusion: Failure."
        </div>

        ```{.py .python linenums="1" title="Example 8: Column missing"}
        >>> keep_first_record_by_columns(df, "f")
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        ColumnDoesNotExistError: Column 'f' does not exist in the DataFrame.
        Try one of: ["a", "b", "c", "d", "e"]
        ```
        !!! failure "Conclusion: Column missing."
        </div>

    ??? info "Notes"
        The way this process will retain only the first record in the given `#!py columns` is by:

        1. Add a new column called `RowNum`
            1. This `RowNum` column uses the SparkSQL function `#!sql ROW_NUMBER()`
            1. The window-function `#!sql OVER` clause will then:
                1. `#!sql PARTITION BY` the `#!py columns`,
                1. `#!sql ORDER BY` the `#!py columns`.
        1. Filter so that `#!sql RowNum=1`.
        1. Drop the `#!py RowNum` column.

    ???+ tip "See Also"
        - [`toolbox_pyspark.checks.assert_columns_exists()`][toolbox_pyspark.checks.assert_columns_exists]
    """
    columns = [columns] if is_type(columns, str) else columns
    assert_columns_exists(dataframe, columns)
    return (
        dataframe.withColumn(
            colName="RowNum",
            col=F.expr(
                f"""
                ROW_NUMBER()
                OVER
                (
                    PARTITION BY {','.join(columns)}
                    ORDER BY {','.join(columns)}
                )
                """
            ),
        )
        .where("RowNum=1")
        .drop("RowNum")
    )


@typechecked
def convert_dataframe(
    dataframe: psDataFrame,
    return_type: Union[
        LITERAL_PYSPARK_DATAFRAME_NAMES,
        LITERAL_PANDAS_DATAFRAME_NAMES,
        LITERAL_NUMPY_ARRAY_NAMES,
        LITERAL_LIST_OBJECT_NAMES,
        str,
    ] = "pd",
) -> Optional[Union[psDataFrame, pdDataFrame, npArray, list]]:
    """
    !!! note "Summary"
        Convert a PySpark DataFrame to the desired return type.

    ???+ abstract "Details"
        This function converts a PySpark DataFrame to one of the supported return types, including:

        PySpark DataFrame:

        <div class="mdx-four-columns" markdown>

        - `#!py "spark.DataFrame"`
        - `#!py "pyspark.DataFrame"`
        - `#!py "pyspark"`
        - `#!py "spark"`
        - `#!py "ps.DataFrame"`
        - `#!py "ps.df"`
        - `#!py "psdf"`
        - `#!py "psDataFrame"`
        - `#!py "psDF"`
        - `#!py "ps"`

        </div>

        Pandas DataFrame:

        <div class="mdx-four-columns" markdown>

        - `#!py "pandas.DataFrame"`
        - `#!py "pandas"`
        - `#!py "pd.DataFrame"`
        - `#!py "pd.df"`
        - `#!py "pddf"`
        - `#!py "pdDataFrame"`
        - `#!py "pdDF"`
        - `#!py "pd"`

        </div>

        NumPy array:

        <div class="mdx-four-columns" markdown>

        - `#!py "numpy.array"`
        - `#!py "np.array"`
        - `#!py "np"`
        - `#!py "numpy"`
        - `#!py "nparr"`
        - `#!py "npa"`
        - `#!py "np.arr"`
        - `#!py "np.a"`

        </div>

        Python list:

        <div class="mdx-four-columns" markdown>

        - `#!py "list"`
        - `#!py "lst"`
        - `#!py "l"`
        - `#!py "flat_list"`
        - `#!py "flatten_list"`

        </div>

    Params:
        dataframe (psDataFrame):
            The PySpark DataFrame to be converted.
        return_type (Union[LITERAL_LIST_OBJECT_NAMES, LITERAL_PANDAS_DATAFRAME_NAMES, LITERAL_PYSPARK_DATAFRAME_NAMES, LITERAL_NUMPY_ARRAY_NAMES, str], optional):
            The desired return type.<br>
            Options:

            - `#!py "ps"`: Return the PySpark DataFrame.
            - `#!py "pd"`: Return a Pandas DataFrame.
            - `#!py "np"`: Return a NumPy array.
            - `#!py "list"`: Return a Python list.
            - `#!py "list_flat"`: Return a flat Python list (1D).

            Defaults to `#!py "pd"` (Pandas DataFrame).

    Raises:
        TypeError:
            If any of the inputs parsed to the parameters of this function are not the correct type. Uses the [`@typeguard.typechecked`](https://typeguard.readthedocs.io/en/stable/api.html#typeguard.typechecked) decorator.
        ValueError:
            If any of the values parsed to `return_type` are not valid options.

    Returns:
        (Optional[Union[psDataFrame, pdDataFrame, npArray, list]]):
            The converted data in the specified return type.

    ???+ example "Examples"

        ```{.py .python linenums="1" title="Set up"}
        >>> # Imports
        >>> import pandas as pd
        >>> from pyspark.sql import SparkSession
        >>> from toolbox_pyspark.cleaning import convert_dataframe
        >>>
        >>> # Instantiate Spark
        >>> spark = SparkSession.builder.getOrCreate()
        >>>
        >>> # Create data
        >>> df = spark.createDataFrame(
        ...     pdDataFrame(
        ...         {
        ...             "a": [1, 2, 3, 4],
        ...             "b": ["a", "b", "c", "d"],
        ...         }
        ...     )
        ... )
        >>>
        >>> # Check
        >>> df.show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+
        | a | b |
        +---+---+
        | 0 | a |
        | 1 | b |
        | 2 | c |
        | 3 | d |
        +---+---+
        ```

        ```{.py .python linenums="1" title="Example 1: Convert to PySpark"}
        >>> new_df = convert_dataframe(df, "ps")
        >>> print(type(new_df))
        >>> new_df.show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        <class 'pyspark.sql.dataframe.DataFrame'>
        ```
        ```{.txt .text title="Terminal"}
        +---+---+
        | a | b |
        +---+---+
        | 0 | a |
        | 1 | b |
        | 2 | c |
        | 3 | d |
        +---+---+
        ```
        !!! success "Conclusion: Successfully converted to PySpark."
        </div>

        ```{.py .python linenums="1" title="Example 2: Convert to Pandas"}
        >>> new_df = convert_dataframe(df, "pd")
        >>> print(type(new_df))
        >>> print(new_df)
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        <class 'pandas.core.frame.DataFrame'>
        ```
        ```{.txt .text title="Terminal"}
           a  b
        0  0  a
        1  1  b
        2  2  c
        3  3  d
        ```
        !!! success "Conclusion: Successfully converted to Pandas."
        </div>

        ```{.py .python linenums="1" title="Example 3: Convert to Numpy"}
        >>> new_df = convert_dataframe(df, "np")
        >>> print(type(new_df))
        >>> print(new_df)
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        <class 'numpy.ndarray'>
        ```
        ```{.txt .text title="Terminal"}
        [[0 "a"]
         [1 "b"]
         [2 "c"]
         [3 "d"]]
        ```
        !!! success "Conclusion: Successfully converted to Numpy."
        </div>

        ```{.py .python linenums="1" title="Example 4: List"}
        >>> new_df = convert_dataframe(df, "list")
        >>> print(type(new_df))
        >>> print(new_df)
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        <class 'list'>
        ```
        ```{.txt .text title="Terminal"}
        [
            [0, "a"],
            [1, "b"],
            [2, "c"],
            [3, "d"],
        ]
        ```
        !!! success "Conclusion: Successfully converted to List."
        </div>

        ```{.py .python linenums="1" title="Example 5: Convert to single column as list"}
        >>> new_df = convert_dataframe(df.select("b"), "flat_list")
        >>> print(type(new_df))
        >>> print(new_df)
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        <class 'list'>
        ```
        ```{.txt .text title="Terminal"}
        ["a", "b", "c", "d"]
        ```
        !!! success "Conclusion: Successfully converted to flat List."
        </div>

        ```{.py .python linenums="1" title="Example 6: Invalid return type"}
        >>> convert_dataframe(df, "invalid")
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        ValueError: Unknown return type: 'invalid'.
        Must be one of: ['pd', 'ps', 'np', 'list'].
        For more info, check the `constants` module.
        ```
        !!! failure "Conclusion: Invalid return type."
        </div>

    ???+ tip "See Also"
        - [`toolbox_pyspark.constants`][toolbox_pyspark.constants]
    """
    if return_type in VALID_PYSPARK_DATAFRAME_NAMES:
        return dataframe
    elif return_type in VALID_PANDAS_DATAFRAME_NAMES:
        return dataframe.toPandas()
    elif return_type in VALID_NUMPY_ARRAY_NAMES:
        return dataframe.toPandas().values  # type:ignore
    elif return_type in VALID_LIST_OBJECT_NAMES:
        if "flat" in return_type:
            return flatten(dataframe.toPandas().values.tolist())  # type:ignore
        else:
            return dataframe.toPandas().values.tolist()  # type:ignore
    else:
        raise ValueError(
            f"Unknown return type: '{return_type}'.\n"
            f"Must be one of: {['pd', 'ps', 'np', 'list']}.\n"
            f"For more info, check the `constants` module."
        )


@typechecked
def get_column_values(
    dataframe: psDataFrame,
    column: str,
    distinct: bool = True,
    return_type: Union[
        LITERAL_PYSPARK_DATAFRAME_NAMES,
        LITERAL_PANDAS_DATAFRAME_NAMES,
        LITERAL_NUMPY_ARRAY_NAMES,
        LITERAL_LIST_OBJECT_NAMES,
        str,
    ] = "pd",
) -> Optional[Union[psDataFrame, pdDataFrame, npArray, list]]:
    """
    !!! note "Summary"
        Extract and return unique values from a specified column of a PySpark DataFrame.

    ???+ abstract "Details"
        This function retrieves the unique values from a specified column of a PySpark DataFrame.

    Params:
        dataframe (psDataFrame):
            The input PySpark DataFrame.
        column (str):
            The name of the column from which to extract unique values.
        distinct (bool, optional):
            If `#!py True`, return distinct (unique) values.<br>
            Defaults to `#!py True`.
        return_type (Union[LITERAL_LIST_OBJECT_NAMES, LITERAL_PANDAS_DATAFRAME_NAMES, LITERAL_PYSPARK_DATAFRAME_NAMES, LITERAL_NUMPY_ARRAY_NAMES, str], optional):
            The desired return type.<br>
            Options:

            - `#!py "ps"`: Return the result as a PySpark DataFrame.
            - `#!py "pd"`: Return the result as a Pandas DataFrame.
            - `#!py "np"`: Return the result as a NumPy array.
            - `#!py "list"`: Return the result as a Python list.

            Defaults to `#!py "pd"`.

    Raises:
        TypeError:
            If any of the inputs parsed to the parameters of this function are not the correct type. Uses the [`@typeguard.typechecked`](https://typeguard.readthedocs.io/en/stable/api.html#typeguard.typechecked) decorator.
        ValueError:
            If any of the values parsed to `return_type` are not valid options.
        ColumnDoesNotExistError:
            If the `#!py column` does not exist within `#!py dataframe.columns`.

    Returns:
        (Optional[Union[psDataFrame, pdDataFrame, npArray, list]]):
            The values from the given column, in the desired type.

    ???+ example "Examples"

        ```{.py .python linenums="1" title="Set up"}
        >>> # Imports
        >>> import pandas as pd
        >>> from pyspark.sql import SparkSession
        >>> from toolbox_pyspark.cleaning import get_column_values
        >>>
        >>> # Instantiate Spark
        >>> spark = SparkSession.builder.getOrCreate()
        >>>
        >>> # Create data
        >>> df = spark.createDataFrame(
        ...     pd.DataFrame(
        ...         {
        ...             "a": [0, 1, 2, 3],
        ...             "b": ["a", "b", "c", "d"],
        ...             "c": ["c", "c", "c", "c"],
        ...             "d": ["d", "d", "d", "d"],
        ...         }
        ...     )
        ... )
        >>>
        >>> # Check
        >>> df.show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+
        | a | b | c | d |
        +---+---+---+---+
        | 0 | a | c | d |
        | 1 | b | c | d |
        | 2 | c | c | d |
        | 3 | d | c | d |
        +---+---+---+---+
        ```
        </div>

        ```{.py .python linenums="1" title="Example 1: Default params"}
        >>> values = get_column_values(df, "c")
        >>> print(type(values))
        >>> print(values)
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        <class 'pandas.core.frame.DataFrame'>
        ```
        ```{.txt .text title="Terminal"}
          c
        0 c
        ```
        !!! success "Conclusion: Successfully extracted distinct values from the `c` column, and converted to Pandas."
        </div>

        ```{.py .python linenums="1" title="Example 2: Not distinct"}
        >>> values = get_column_values(df, "c", False)
        >>> print(type(values))
        >>> print(values)
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        <class 'pandas.core.frame.DataFrame'>
        ```
        ```{.txt .text title="Terminal"}
          c
        0 c
        1 c
        2 c
        3 c
        ```
        !!! success "Conclusion: Successfully extracted values from the `c` column, and converted to Pandas."
        </div>

        ```{.py .python linenums="1" title="Example 3: Flat list"}
        >>> values = get_column_values(df, "c", False, "flat_list")
        >>> print(type(values))
        >>> print(values)
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        <class 'list'>
        ```
        ```{.txt .text title="Terminal"}
        ["c", "c", "c", "c"]
        ```
        !!! success "Conclusion: Successfully extracted values from the `c` column, and converted to flat List."
        </div>

        ```{.py .python linenums="1" title="Example 4: Invalid return type"}
        >>> get_column_values(df, "c", return_type="invalid")
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        ValueError: Unknown return type: 'invalid'.
        Must be one of: ['pd', 'ps', 'np', 'list'].
        For more info, check the `constants` module.
        ```
        !!! failure "Conclusion: Invalid return type."
        </div>

    ???+ tip "See Also"
        - [`toolbox_pyspark.cleaning.convert_dataframe()`][toolbox_pyspark.cleaning.convert_dataframe]
        - [`toolbox_pyspark.constants`][toolbox_pyspark.constants]
    """
    df: psDataFrame = dataframe.select(column).filter(
        f"{column} is not null and {column} <> ''"
    )
    df = df.distinct() if distinct else df
    return convert_dataframe(dataframe=df, return_type=return_type)


@typechecked
def update_nullability(
    dataframe: psDataFrame,
    columns: Optional[Union[str, str_collection]] = None,
    nullable: bool = True,
) -> psDataFrame:
    """
    !!! note "Summary"
        Update the nullability of specified columns in a PySpark DataFrame.

    ???+ abstract "Details"
        This function updates the nullability of the specified columns in a PySpark DataFrame. If no columns are specified, it updates the nullability of all columns.

    Params:
        dataframe (psDataFrame):
            The input PySpark DataFrame.
        columns (Optional[Union[str, str_collection]], optional):
            The columns for which to update nullability. If not provided, all columns will be updated.<br>
            Defaults to `#!py None`.
        nullable (bool, optional):
            Whether to set the columns as nullable or not.<br>
            Defaults to `#!py True`.

    Raises:
        TypeError:
            If any of the inputs parsed to the parameters of this function are not the correct type. Uses the [`@typeguard.typechecked`](https://typeguard.readthedocs.io/en/stable/api.html#typeguard.typechecked) decorator.
        ColumnDoesNotExistError:
            If any of the `#!py columns` do not exist within `#!py dataframe.columns`.

    Returns:
        (psDataFrame):
            The updated DataFrame with the specified columns' nullability updated.

    ???+ example "Examples"

        ```{.py .python linenums="1" title="Set up"}
        >>> # Imports
        >>> import pandas as pd
        >>> from pyspark.sql import SparkSession
        >>> from toolbox_pyspark.cleaning import update_nullability
        >>>
        >>> # Instantiate Spark
        >>> spark = SparkSession.builder.getOrCreate()
        >>>
        >>> # Create data
        >>> df = spark.createDataFrame(
        ...     pd.DataFrame(
        ...         {
        ...             "a": [1, 2, 3, 4],
        ...             "b": ["a", "b", "c", "d"],
        ...             "c": [1.1, 2.2, 3.3, 4.4],
        ...         }
        ...     )
        ... )
        >>>
        >>> # Check
        >>> df.show()
        >>> print(df.schema)
        >>> df.printSchema()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+-----+
        | a | b |   c |
        +---+---+-----+
        | 1 | a | 1.1 |
        | 2 | b | 2.2 |
        | 3 | c | 3.3 |
        | 4 | d | 4.4 |
        +---+---+-----+
        ```
        ```{.txt .text title="Terminal"}
        StructType(
            [
                StructField("a", LongType(), True),
                StructField("b", StringType(), True),
                StructField("c", DoubleType(), True),
            ]
        )
        ```
        ```{.txt .text title="Terminal"}
        root
         |-- a: long (nullable = true)
         |-- b: string (nullable = true)
         |-- c: double (nullable = true)
        ```
        </div>

        ```{.py .python linenums="1" title="Example 1: Update nullability of all columns"}
        >>> new_df = update_nullability(df, nullable=False)
        >>> new_df.printSchema()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        root
         |-- a: long (nullable = false)
         |-- b: string (nullable = false)
         |-- c: double (nullable = false)
        ```
        !!! success "Conclusion: Successfully updated nullability of all columns."
        </div>

        ```{.py .python linenums="1" title="Example 2: Update nullability of specific columns"}
        >>> new_df = update_nullability(df, columns=["a", "c"], nullable=False)
        >>> new_df.printSchema()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        root
         |-- a: long (nullable = false)
         |-- b: string (nullable = true)
         |-- c: double (nullable = false)
        ```
        !!! success "Conclusion: Successfully updated nullability of specific columns."
        </div>

        ```{.py .python linenums="1" title="Example 3: Column does not exist"}
        >>> update_nullability(df, columns="d", nullable=False)
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        ColumnDoesNotExistError: Column 'd' does not exist in the DataFrame.
        Try one of: ["a", "b", "c"]
        ```
        !!! failure "Conclusion: Column does not exist."
        </div>

    ??? success "Credit"
        All credit goes to: https://stackoverflow.com/questions/46072411/can-i-change-the-nullability-of-a-column-in-my-spark-dataframe#answer-51821437.

    ???+ tip "See Also"
        - [`toolbox_pyspark.checks.assert_columns_exists()`][toolbox_pyspark.checks.assert_columns_exists]
    """
    columns = get_columns(dataframe, columns)
    assert_columns_exists(dataframe=dataframe, columns=columns)
    schema: T.StructType = dataframe.schema
    for struct_field in schema:
        if struct_field.name in columns:
            struct_field.nullable = nullable
    return dataframe.sparkSession.createDataFrame(
        data=dataframe.rdd, schema=dataframe.schema
    )


# ---------------------------------------------------------------------------- #
#  Trimming                                                                 ####
# ---------------------------------------------------------------------------- #


@typechecked
def trim_spaces_from_column(
    dataframe: psDataFrame,
    column: str,
) -> psDataFrame:
    """
    !!! note "Summary"
        For a given list of columns, trim all of the excess white spaces from them.

    Params:
        dataframe (psDataFrame):
            The DataFrame to update.
        column (str):
            The column to clean.

    Raises:
        TypeError:
            If any of the inputs parsed to the parameters of this function are not the correct type. Uses the [`@typeguard.typechecked`](https://typeguard.readthedocs.io/en/stable/api.html#typeguard.typechecked) decorator.
        ColumnDoesNotExistError:
            If the `#!py column` does not exist within `#!py dataframe.columns`.

    Returns:
        (psDataFrame):
            The updated Data Frame.

    ???+ example "Examples"

        ```{.py .python linenums="1" title="Set up"}
        >>> # Imports
        >>> import pandas as pd
        >>> from pyspark.sql import SparkSession
        >>> from toolbox_pyspark.cleaning import trim_spaces_from_column
        >>>
        >>> # Instantiate Spark
        >>> spark = SparkSession.builder.getOrCreate()
        >>>
        >>> # Create data
        >>> df = spark.createDataFrame(
        ...     pd.DataFrame(
        ...         {
        ...             "a": [1, 2, 3, 4],
        ...             "b": ["a", "b", "c", "d"],
        ...             "c": ["1   ", "1   ", "1   ", "1   "],
        ...             "d": ["   2", "   2", "   2", "   2"],
        ...             "e": ["   3   ", "   3   ", "   3   ", "   3   "],
        ...         }
        ...     )
        ... )
        >>>
        >>> # Check
        ```{.py .python linenums="1" title="Check"}
        >>> df.show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+------+------+---------+
        | a | b |    c |    d |       e |
        +---+---+------+------+---------+
        | 1 | a | 1    |    2 |    3    |
        | 2 | b | 1    |    2 |    3    |
        | 3 | c | 1    |    2 |    3    |
        | 4 | d | 1    |    2 |    3    |
        +---+---+------+------+---------+
        ```
        </div>

        ```{.py .python linenums="1" title="Example 1: Trim column"}
        >>> trim_spaces_from_column(df, "c").show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+------+--------+
        | a | b | c |    d |      e |
        +---+---+---+------+--------+
        | 1 | a | 1 |    2 |   2    |
        | 2 | b | 1 |    2 |   2    |
        | 3 | c | 1 |    2 |   2    |
        | 4 | d | 1 |    2 |   2    |
        +---+---+---+------+--------+
        ```
        !!! success "Conclusion: Successfully trimmed the `c` column."
        </div>

        ```{.py .python linenums="1" title="Example 2: Invalid column"}
        >>> trim_spaces_from_column(df, "f")
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        ColumnDoesNotExistError: Column 'f' does not exist in the DataFrame.
        Try one of: ["a", "b", "c", "d", "e"]
        ```
        !!! failure "Conclusion: Column does not exist."
        </div>

    ??? info "Notes"

        ???+ info "Justification"
            - The main reason for this function is because when the data was exported from the Legacy WMS's, there's a _whole bunch_ of trailing spaces in the data fields. My theory is because of the data type in the source system. That is, if it's originally stored as 'char' type, then it will maintain the data length. This issues doesn't seem to be affecting the `varchar` fields. Nonetheless, this function will strip the white spaces from the data; thus reducing the total size of the data stored therein.
            - The reason why it is necessary to write this out as a custom function, instead of using the [`F.trim()`][trim] function from the PySpark library directly is due to the deficiencies of the Java [`trim()`](https://docs.oracle.com/javase/8/docs/api/java/lang/String.html#trim) function. More specifically, there are 13 different whitespace characters available in our ascii character set. The Java function only cleans about 6 of these. So therefore, we define this function which iterates through all 13 whitespace characters, and formats them in to a regular expression, to then parse it to the [`F.regexp_replace()`][regexp_replace] function to be replaced with an empty string (`""`). Therefore, all 13 characters will be replaced, the strings will be cleaned and trimmed ready for further processing.

            [trim]: https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.functions.trim.html
            [regexp_replace]: https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.functions.regexp_replace.html

        ???+ info "Regex definition: `^[...]+|[...]+$`"
            - 1st Alternative: '^[...]+'
                - '^' asserts position at start of a line
                - Match a single character present in the list below '[...]'
                    - '+' matches the previous token between one and unlimited times, as many times as possible, giving back as needed (greedy)
                    - matches a single character in the list '  ' (case sensitive)
                        - matches the character ' ' with index 160 (A0 or 240) literally (case sensitive)
                        - matches the character ' ' with index 32 (20 or 40) literally (case sensitive)
                        - ... (repeat for all whitespace characters)
            - 2nd Alternative: '[...]+$'
                - Match a single character present in the list below '[...]'
                    - '+' matches the previous token between one and unlimited times, as many times as possible, giving back as needed (greedy)
                    - matches a single character in the list '  ' (case sensitive)
                        - matches the character ' ' with index 160 (A0 or 240) literally (case sensitive)
                        - matches the character ' ' with index 32 (20 or 40) literally (case sensitive)
                        - ... (repeat for all whitespace characters)
                - '$' asserts position at the end of a line

    ??? tip "See Also"
        - [`trim_spaces_from_columns()`][toolbox_pyspark.cleaning.trim_spaces_from_columns]
        - [`ALL_WHITESPACE_CHARACTERS`][toolbox_pyspark.constants.ALL_WHITESPACE_CHARACTERS]
    """
    assert_column_exists(dataframe=dataframe, column=column, match_case=True)
    space_chars: str_list = [chr(char.ascii) for char in WHITESPACES]
    regexp: str = f"^[{''.join(space_chars)}]+|[{''.join(space_chars)}]+$"
    return dataframe.withColumn(column, F.regexp_replace(column, regexp, ""))


@typechecked
def trim_spaces_from_columns(
    dataframe: psDataFrame,
    columns: Optional[Union[str, str_collection]] = None,
) -> psDataFrame:
    """
    !!! note "Summary"
        For a given list of columns, trim all of the excess white spaces from them.

    Params:
        dataframe (psDataFrame):
            The DataFrame to be updated.
        columns (Optional[Union[str, str_collection]], optional):
            The list of columns to be updated.
            Must be valid columns on `dataframe`.
            If given as a string, will be executed as a single column (ie. one-element long list).
            If not given, will apply to all columns in `dataframe` which have the data-type `string`.
            It is also possible to parse the values `#!py "all"` or `#!py "all_string"`, which will also apply this function to all columns in `dataframe` which have the data-type `string`.<br>
            Defaults to `#!py None`.

    Returns:
        (psDataFrame):
            The updated DataFrame.

    ???+ example "Examples"

        ```{.py .python linenums="1" title="Set up"}
        >>> # Imports
        >>> import pandas as pd
        >>> from pyspark.sql import SparkSession
        >>> from toolbox_pyspark.cleaning import trim_spaces_from_columns
        >>>
        >>> # Instantiate Spark
        >>> spark = SparkSession.builder.getOrCreate()
        >>>
        >>> # Create data
        >>> df = spark.createDataFrame(
        ...     pd.DataFrame(
        ...         {
        ...             "a": [1, 2, 3, 4],
        ...             "b": ["a", "b", "c", "d"],
        ...             "c": ["1   ", "1   ", "1   ", "1   "],
        ...             "d": ["   2", "   2", "   2", "   2"],
        ...             "e": ["   3   ", "   3   ", "   3   ", "   3   "],
        ...         }
        ...     )
        ... )
        >>>
        >>> # Check
        ```{.py .python linenums="1" title="Check"}
        >>> df.show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+------+------+---------+
        | a | b |    c |    d |       e |
        +---+---+------+------+---------+
        | 1 | a | 1    |    2 |    3    |
        | 2 | b | 1    |    2 |    3    |
        | 3 | c | 1    |    2 |    3    |
        | 4 | d | 1    |    2 |    3    |
        +---+---+------+------+---------+
        ```
        </div>

        ```{.py .python linenums="1" title="Example 1: One column as list"}
        >>> trim_spaces_from_columns(df, ["c"]).show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+------+---------+
        | a | b | c |    d |       e |
        +---+---+---+------+---------+
        | 1 | a | 1 |    2 |    3    |
        | 2 | b | 1 |    2 |    3    |
        | 3 | c | 1 |    2 |    3    |
        | 4 | d | 1 |    2 |    3    |
        +---+---+---+------+---------+
        ```
        !!! success "Conclusion: Successfully trimmed the `c` column."
        </div>

        ```{.py .python linenums="1" title="Example 2: Single column as string"}
        >>> trim_spaces_from_columns(df, "d").show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+------+---+---------+
        | a | b |    c | d |       e |
        +---+---+------+---+---------+
        | 1 | a | 1    | 2 |    3    |
        | 2 | b | 1    | 2 |    3    |
        | 3 | c | 1    | 2 |    3    |
        | 4 | d | 1    | 2 |    3    |
        +---+---+------+---+---------+
        ```
        !!! success "Conclusion: Successfully trimmed the `d` column."
        </div>

        ```{.py .python linenums="1" title="Example 3: Multiple columns"}
        >>> trim_spaces_from_columns(df, ["c", "d"]).show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+---------+
        | a | b | c | d |       e |
        +---+---+---+---+---------+
        | 1 | a | 1 | 2 |    3    |
        | 2 | b | 1 | 2 |    3    |
        | 3 | c | 1 | 2 |    3    |
        | 4 | d | 1 | 2 |    3    |
        +---+---+---+---+---------+
        ```
        !!! success "Conclusion: Successfully trimmed the `c` and `d` columns."
        </div>

        ```{.py .python linenums="1" title="Example 4: All columns"}
        >>> trim_spaces_from_columns(df, "all").show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+---+
        | a | b | c | d | e |
        +---+---+---+---+---+
        | 1 | a | 1 | 2 | 3 |
        | 2 | b | 1 | 2 | 3 |
        | 3 | c | 1 | 2 | 3 |
        | 4 | d | 1 | 2 | 3 |
        +---+---+---+---+---+
        ```
        !!! success "Conclusion: Successfully trimmed all columns."
        </div>

        ```{.py .python linenums="1" title="Example 5: Default config"}
        >>> trim_spaces_from_columns(df).show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+---+
        | a | b | c | d | e |
        +---+---+---+---+---+
        | 1 | a | 1 | 2 | 3 |
        | 2 | b | 1 | 2 | 3 |
        | 3 | c | 1 | 2 | 3 |
        | 4 | d | 1 | 2 | 3 |
        +---+---+---+---+---+
        ```
        !!! success "Conclusion: Successfully trimmed all columns."
        </div>

        ```{.py .python linenums="1" title="Example 6: Invalid column"}
        >>> trim_spaces_from_columns(df, ["f"])
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        ColumnDoesNotExistError: Columns ['f'] do not exist in the DataFrame.
        Try one of: ["a", "b", "c", "d", "e"]
        ```
        !!! failure "Conclusion: Columns do not exist."
        </div>

    ???+ info "Notes"

        ???+ info "Justification"
            - The main reason for this function is because when the data was exported from the Legacy WMS's, there's a _whole bunch_ of trailing spaces in the data fields. My theory is because of the data type in the source system. That is, if it's originally stored as 'char' type, then it will maintain the data length. This issues doesn't seem to be affecting the `varchar` fields. Nonetheless, this function will strip the white spaces from the data; thus reducing the total size of the data stored therein.
            - The reason why it is necessary to write this out as a custom function, instead of using the [`F.trim()`][trim] function from the PySpark library directly is due to the deficiencies of the Java [`trim()`](https://docs.oracle.com/javase/8/docs/api/java/lang/String.html#trim) function. More specifically, there are 13 different whitespace characters available in our ascii character set. The Java function only cleans about 6 of these. So therefore, we define this function which iterates through all 13 whitespace characters, and formats them in to a regular expression, to then parse it to the [`F.regexp_replace()`][regexp_replace] function to be replaced with an empty string (`""`). Therefore, all 13 characters will be replaced, the strings will be cleaned and trimmed ready for further processing.
            - The reason why this function exists as a standalone, and does not call [`trim_spaces_from_column()`][toolbox_pyspark.cleaning.trim_spaces_from_column] from within a loop is because [`trim_spaces_from_column()`][toolbox_pyspark.cleaning.trim_spaces_from_column] utilises the [`.withColumn()`][withColumn] method to implement the [`F.regexp_replace()`][regexp_replace] function on columns individually. When implemented iteratively, this process will create huge DAG's for the RDD, and blow out the complexity to a huge extend. Whereas this [`trim_spaces_from_columns()`][toolbox_pyspark.cleaning.trim_spaces_from_columns] function will utilise the [`.withColumns()`][withColumns] method to implement the [`F.regexp_replace()`][regexp_replace] function over all columns at once. This [`.withColumns()`][withColumns] method projects the function down to the underlying dataset in one single execution; not a different execution per column. Therefore, it is more simpler and more efficient.

            [withColumn]: https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrame.withColumn.html
            [withColumns]: https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrame.withColumns.html
            [trim]: https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.functions.trim.html
            [regexp_replace]: https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.functions.regexp_replace.html

        ???+ info "Regex definition: `^[...]+|[...]+$`"
            - 1st Alternative: `^[...]+`
                - `^` asserts position at start of a line
                - Match a single character present in the list below `[...]`
                    - `+` matches the previous token between one and unlimited times, as many times as possible, giving back as needed (greedy)
                    - matches a single character in the list `  ` (case sensitive)
                        - matches the character ` ` with index 160 (A0 or 240) literally (case sensitive)
                        - matches the character ` ` with index 32 (20 or 40) literally (case sensitive)
                        - ... (repeat for all whitespace characters)
            - 2nd Alternative: `[...]+$`
                - Match a single character present in the list below `[...]`
                    - `+` matches the previous token between one and unlimited times, as many times as possible, giving back as needed (greedy)
                    - matches a single character in the list `  ` (case sensitive)
                        - matches the character ` ` with index 160 (A0 or 240) literally (case sensitive)
                        - matches the character ` ` with index 32 (20 or 40) literally (case sensitive)
                        - ... (repeat for all whitespace characters)

    ??? tip "See Also"
        - [`trim_spaces_from_column()`][toolbox_pyspark.cleaning.trim_spaces_from_column]
        - [`ALL_WHITESPACE_CHARACTERS`][toolbox_pyspark.constants.ALL_WHITESPACE_CHARACTERS]
    """
    columns = get_columns(dataframe, columns)
    assert_columns_exists(dataframe=dataframe, columns=columns, match_case=True)
    space_chars: str_list = WHITESPACES.to_list("chr")  # type:ignore
    regexp: str = f"^[{''.join(space_chars)}]+|[{''.join(space_chars)}]+$"
    cols_exprs: dict[str, Column] = {
        col: F.regexp_replace(col, regexp, "") for col in columns
    }
    return dataframe.withColumns(cols_exprs)


# ---------------------------------------------------------------------------- #
#  Applying functions                                                       ####
# ---------------------------------------------------------------------------- #


@typechecked
def apply_function_to_column(
    dataframe: psDataFrame,
    column: str,
    function: str = "upper",
    *function_args,
    **function_kwargs,
) -> psDataFrame:
    """
    !!! note "Summary"
        Apply a given PySpark `function` to a single `column` on `dataframe`.

    ???+ abstract "Details"
        Under the hood, this function will simply call the [`.withColumn()`](https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrame.withColumn.html) method to apply the function named in `function` from the PySpark [`functions`](https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/functions.html) module.
        ```py
        return dataframe.withColumn(column, getattr(F, function)(column, *function_args, **function_kwargs))
        ```

    Params:
        dataframe (psDataFrame):
            The DataFrame to update.
        column (str):
            The column to update.
        function (str, optional):
            The function to execute. Must be a valid function from the PySpark [`functions`](https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/functions.html) module.<br>
            Defaults to `#!py "upper"`.
        *function_args (Any, optional):
            The arguments to push down to the underlying `function`.
        **function_kwargs (Any, optional):
            The keyword arguments to push down to the underlying `function`.

    Returns:
        (psDataFrame):
            The updated DataFrame.

    ???+ example "Examples"

        ```{.py .python linenums="1" title="Set up"}
        >>> # Imports
        >>> import pandas as pd
        >>> from pyspark.sql import SparkSession
        >>> from toolbox_pyspark.cleaning import apply_function_to_column
        >>>
        >>> # Instantiate Spark
        >>> spark = SparkSession.builder.getOrCreate()
        >>>
        >>> # Create data
        >>> df = spark.createDataFrame(
        ...     pd.DataFrame(
        ...         {
        ...             "a": [0, 1, 2, 3],
        ...             "b": ["a", "b", "c", "d"],
        ...             "c": ["c", "c", "c", "c"],
        ...             "d": ["d", "d", "d", "d"],
        ...         }
        ...     )
        ... )
        >>>
        >>> # Check
        ```{.py .python linenums="1" title="Check"}
        >>> df.show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+
        | a | b | c | d |
        +---+---+---+---+
        | 0 | a | c | d |
        | 1 | b | c | d |
        | 2 | c | c | d |
        | 3 | d | c | d |
        +---+---+---+---+
        ```
        </div>

        ```{.py .python linenums="1" title="Example 1: Default params"}
        >>> apply_function_to_column(df, "c").show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+
        | a | b | c | d |
        +---+---+---+---+
        | 0 | a | C | d |
        | 1 | b | C | d |
        | 2 | c | C | d |
        | 3 | d | C | d |
        +---+---+---+---+
        ```
        !!! success "Conclusion: Successfully applied the `upper` function to the `c` column."
        </div>

        ```{.py .python linenums="1" title="Example 2: Simple function"}
        >>> apply_function_to_column(df, "c", "lower").show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+
        | a | b | c | d |
        +---+---+---+---+
        | 0 | a | c | d |
        | 1 | b | c | d |
        | 2 | c | c | d |
        | 3 | d | c | d |
        +---+---+---+---+
        ```
        !!! success "Conclusion: Successfully applied the `lower` function to the `c` column."
        </div>

        ```{.py .python linenums="1" title="Example 3: Complex function, using args"}
        >>> apply_function_to_column(df, "d", "lpad", 5, "?").show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+-------+
        | a | b | c |     d |
        +---+---+---+-------+
        | 0 | a | c | ????d |
        | 1 | b | c | ????d |
        | 2 | c | c | ????d |
        | 3 | d | c | ????d |
        +---+---+---+-------+
        ```
        !!! success "Conclusion: Successfully applied the `lpad` function to the `d` column."
        </div>

        ```{.py .python linenums="1" title="Example 4: Complex function, using kwargs"}
        >>> new_df = apply_function_to_column(
        ...     dataframe=df,
        ...     column="d",
        ...     function="lpad",
        ...     len=5,
        ...     pad="?",
        ... ).show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+-------+
        | a | b | c |     d |
        +---+---+---+-------+
        | 0 | a | c | ????d |
        | 1 | b | c | ????d |
        | 2 | c | c | ????d |
        | 3 | d | c | ????d |
        +---+---+---+-------+
        ```
        !!! success "Conclusion: Successfully applied the `lpad` function to the `d` column."
        </div>

        ```{.py .python linenums="1" title="Example 5: Different complex function, using kwargs"}
        >>> new_df = apply_function_to_column(
        ...     dataframe=df,
        ...     column="b",
        ...     function="regexp_replace",
        ...     pattern="c",
        ...     replacement="17",
        ... ).show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+----+---+---+
        | a |  b | c | d |
        +---+----+---+---+
        | 0 |  a | c | d |
        | 1 |  b | c | d |
        | 2 | 17 | c | d |
        | 3 |  d | c | d |
        +---+----+---+---+
        ```
        !!! success "Conclusion: Successfully applied the `regexp_replace` function to the `b` column."
        </div>

        ```{.py .python linenums="1" title="Example 6: Part of pipe"}
        >>> new_df = df.transform(
        ...     func=apply_function_to_column,
        ...     column="d",
        ...     function="lpad",
        ...     len=5,
        ...     pad="?",
        ... ).show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+-------+
        | a | b | c |     d |
        +---+---+---+-------+
        | 0 | a | c | ????d |
        | 1 | b | c | ????d |
        | 2 | c | c | ????d |
        | 3 | d | c | ????d |
        +---+---+---+-------+
        ```
        !!! success "Conclusion: Successfully applied the `lpad` function to the `d` column."
        </div>

        ```{.py .python linenums="1" title="Example 7: Column name in different case"}
        >>> new_df = df.transform(
        ...     func=apply_function_to_column,
        ...     column="D",
        ...     function="upper",
        ... ).show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+
        | a | b | c | d |
        +---+---+---+---+
        | 0 | a | c | D |
        | 1 | b | c | D |
        | 2 | c | c | D |
        | 3 | d | c | D |
        +---+---+---+---+
        ```
        !!! success "Conclusion: Successfully applied the `upper` function to the `D` column."
        </div>

        ```{.py .python linenums="1" title="Example 8: Invalid column"}
        >>> apply_function_to_column(df, "f")
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        ColumnDoesNotExistError: Column 'f' does not exist in the DataFrame.
        Try one of: ["a", "b", "c", "d"]
        ```
        !!! failure "Conclusion: Column does not exist."
        </div>

    ??? info "Notes"
        - We have to name the `function` parameter as the full name because when this function is executed as part of a chain (by using the PySpark [`.transform()`](https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrame.transform.html) method), that one uses the `func` parameter.

    ??? tip "See Also"
        - [`apply_function_to_columns()`][toolbox_pyspark.cleaning.apply_function_to_columns]
    """
    assert_column_exists(dataframe, column, False)
    return dataframe.withColumn(
        colName=column,
        col=getattr(F, function)(column, *function_args, **function_kwargs),
    )


@typechecked
def apply_function_to_columns(
    dataframe: psDataFrame,
    columns: Union[str, str_collection],
    function: str = "upper",
    *function_args,
    **function_kwargs,
) -> psDataFrame:
    """
    !!! note "Summary"
        Apply a given PySpark `function` over multiple `columns` on a given `dataframe`.

    ???+ abstract "Details"
        Under the hood, this function will simply call the [`.withColumns()`](https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrame.withColumns.html) method to apply the function named in `function` from the PySpark [`functions`](https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/functions.html) module.
        ```py
        return dataframe.withColumns(
            {column: getattr(F, function)(column, *args, **kwargs) for column in columns}
        )
        ```

    Params:
        dataframe (psDataFrame):
            The DataFrame to update.
        columns (Union[str, str_collection]):
            The columns to update.
        function (str, optional):
            The function to use.<br>
            Defaults to `#!py "upper"`.

    Returns:
        (psDataFrame):
            The updated DataFrame.

    ???+ example "Examples"

        ```{.py .python linenums="1" title="Set up"}
        >>> # Imports
        >>> import pandas as pd
        >>> from pyspark.sql import SparkSession
        >>> from toolbox_pyspark.cleaning import apply_function_to_columns
        >>>
        >>> # Instantiate Spark
        >>> spark = SparkSession.builder.getOrCreate()
        >>>
        >>> # Create data
        >>> df = spark.createDataFrame(
        ...     pd.DataFrame(
        ...         {
        ...             "a": [0, 1, 2, 3],
        ...             "b": ["a", "b", "c", "d"],
        ...             "c": ["c", "c", "c", "c"],
        ...             "d": ["d", "d", "d", "d"],
        ...         }
        ...     )
        ... )
        >>>
        >>> # Check
        ```{.py .python linenums="1" title="Check"}
        >>> df.show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+
        | a | b | c | d |
        +---+---+---+---+
        | 0 | a | c | d |
        | 1 | b | c | d |
        | 2 | c | c | d |
        | 3 | d | c | d |
        +---+---+---+---+
        ```
        </div>

        ```{.py .python linenums="1" title="Example 1: Default params"}
        >>> apply_function_to_columns(df, ["b", "c"]).show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+
        | a | b | c | d |
        +---+---+---+---+
        | 0 | A | C | d |
        | 1 | B | C | d |
        | 2 | C | C | d |
        | 3 | D | C | d |
        +---+---+---+---+
        ```
        !!! success "Conclusion: Successfully applied the `upper` function to the `b` and `c` columns."
        </div>

        ```{.py .python linenums="1" title="Example 2: Simple function"}
        >>> apply_function_to_columns(df, ["b", "c"], "lower").show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+
        | a | b | c | d |
        +---+---+---+---+
        | 0 | a | c | d |
        | 1 | b | c | d |
        | 2 | c | c | d |
        | 3 | d | c | d |
        +---+---+---+---+
        ```
        !!! success "Conclusion: Successfully applied the `lower` function to the `b` and `c` columns."
        </div>

        ```{.py .python linenums="1" title="Example 3: Complex function, with args"}
        >>> apply_function_to_columns(df, ["b", "c", "d"], "lpad", 5, "?").show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+-------+-------+-------+
        | a |     b |     c |     d |
        +---+-------+-------+-------+
        | 0 | ????a | ????c | ????d |
        | 1 | ????b | ????c | ????d |
        | 2 | ????c | ????c | ????d |
        | 3 | ????d | ????c | ????d |
        +---+-------+-------+-------+
        ```
        !!! success "Conclusion: Successfully applied the `lpad` function to the `b`, `c` and `d` columns."
        </div>

        ```{.py .python linenums="1" title="Example 4: Complex function, with kwargs"}
        >>> apply_function_to_columns(
        ...     dataframe=df,
        ...     columns=["b", "c", "d"],
        ...     function="lpad",
        ...     len=5,
        ...     pad="?",
        ... ).show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+-------+-------+-------+
        | a |     b |     c |     d |
        +---+-------+-------+-------+
        | 0 | ????a | ????c | ????d |
        | 1 | ????b | ????c | ????d |
        | 2 | ????c | ????c | ????d |
        | 3 | ????d | ????c | ????d |
        +---+-------+-------+-------+
        ```
        !!! success "Conclusion: Successfully applied the `lpad` function to the `b`, `c` and `d` columns."
        </div>

        ```{.py .python linenums="1" title="Example 5: Different complex function, with kwargs"}
        >>> apply_function_to_columns(
        ...     dataframe=df,
        ...     columns=["b", "c", "d"],
        ...     function="regexp_replace",
        ...     pattern="c",
        ...     replacement="17",
        ... ).show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+----+----+---+
        | a |  b |  c | d |
        +---+----+----+---+
        | 0 |  a | 17 | d |
        | 1 |  b | 17 | d |
        | 2 | 17 | 17 | d |
        | 3 |  d | 17 | d |
        +---+----+----+---+
        ```
        !!! success "Conclusion: Successfully applied the `regexp_replace` function to the `b`, `c` and `d` columns."
        </div>

        ```{.py .python linenums="1" title="Example 6: Part of pipe"}
        >>> df.transform(
        ...     func=apply_function_to_columns,
        ...     columns=["b", "c", "d"],
        ...     function="lpad",
        ...     len=5,
        ...     pad="?",
        ... ).show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+-------+-------+-------+
        | a |     b |     c |     d |
        +---+-------+-------+-------+
        | 0 | ????a | ????c | ????d |
        | 1 | ????b | ????c | ????d |
        | 2 | ????c | ????c | ????d |
        | 3 | ????d | ????c | ????d |
        +---+-------+-------+-------+
        ```
        !!! success "Conclusion: Successfully applied the `lpad` function to the `b`, `c` and `d` columns."
        </div>

        ```{.py .python linenums="1" title="Example 7: Column name in different case"}
        >>> apply_function_to_columns(
        ...     dataframe=df,
        ...     columns=["B", "c", "D"],
        ...     function="upper",
        ... ).show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+
        | a | b | c | d |
        +---+---+---+---+
        | 0 | A | C | D |
        | 1 | B | C | D |
        | 2 | C | C | D |
        | 3 | D | C | D |
        +---+---+---+---+
        ```
        !!! success "Conclusion: Successfully applied the `upper` function to the `B`, `c` and `D` columns."
        </div>

        ```{.py .python linenums="1" title="Example 8: Invalid columns"}
        >>> apply_function_to_columns(df, ["f"])
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        ColumnDoesNotExistError: Columns ['f'] do not exist in the DataFrame.
        Try one of: ["a", "b", "c", "d"]
        ```
        !!! failure "Conclusion: Columns do not exist."
        </div>

    ??? info "Notes"
        - We have to name the `function` parameter as the full name because when this function is executed as part of a chain (by using the PySpark [`.transform()`](https://spark.apache.org/docs/latest/api/python/reference/pyspark.sql/api/pyspark.sql.DataFrame.transform.html) method), that one uses the `func` parameter.

    ??? tip "See Also"
        - [`apply_function_to_column()`][toolbox_pyspark.cleaning.apply_function_to_column]
    """
    columns = get_columns(dataframe, columns)
    assert_columns_exists(dataframe, columns, False)
    return dataframe.withColumns(
        {
            column: getattr(F, function)(column, *function_args, **function_kwargs)
            for column in columns
        }
    )


# ---------------------------------------------------------------------------- #
#  Clean across tables                                                      ####
# ---------------------------------------------------------------------------- #


@typechecked
def drop_matching_rows(
    left_table: psDataFrame,
    right_table: psDataFrame,
    on_keys: Union[str, str_collection],
    join_type: VALID_PYAPARK_JOIN_TYPES = "left_anti",
    where_clause: Optional[str] = None,
) -> psDataFrame:
    """
    !!! note "Summary"
        This function is designed to _remove_ any rows on the `left_table` which _are_ existing on the `right_table`. That's why the `join_type` should always be `left_anti`.

    ???+ abstract "Details"
        The intention behind this function is originating from the `Accumulation` layer in the BigDaS environment. The process on this table layer is to only _insert_ rows from the `left_table` to the `right_table` with are **not existing** on the `right_table`. We include the `where_clause` here so that we can control any updated rows. Specifically, we check the `editdatetime` field between the `left_table` and the `right_table`, and any record on the `left_table` where the `editdatetime` field is _greater than_ the `editdatetime` value on the `right_table`, then this row will _remain_ on the `left_table`, and will later be _updated_ on the `right_table`.

        It's important to specify here that this function was created to handle the _same table_ between the `left_table` and the `right_table`, which are existing between different layers in the ADLS environment. Logically, it can be used for other purposes (it's generic enough); however, the intention was specifically for cleaning during the data pipeline processes.

    Params:
        left_table (psDataFrame):
            The DataFrame _from which_ you will be deleting the records.
        right_table (psDataFrame):
            The DataFrame _from which_ to check for existing records. If any matching `on_keys` are existing on both the `right_table` and the `left_table`, then those records will be deleted from the `left_table`.
        on_keys (Union[str, str_collection]):
            The matching keys between the two tables. These keys (aka columns) must be existing on both the `left_table` and the `right_table`.
        join_type (VALID_PYAPARK_JOIN_TYPES, optional):
            The type of join to use for this process. For the best performance, keep it as the default value.<br>
            Defaults to `#!py "left_anti"`.
        where_clause (Optional[str], optional):
            Any additional conditions to place on this join. Any records which **match** this condition will be **kept** on the `left_table`.<br>
            Defaults to `#!py None`.

    Returns:
        (psDataFrame):
            The `left_table` after it has had it's rows deleted and cleaned by the `right_table`.

    ???+ example "Examples"

        ```{.py .python linenums="1" title="Set up"}
        >>> # Imports
        >>> import pandas as pd
        >>> from pyspark.sql import SparkSession
        >>> from toolbox_pyspark.cleaning import drop_matching_rows
        >>>
        >>> # Instantiate Spark
        >>> spark = SparkSession.builder.getOrCreate()
        >>>
        >>> # Create data
        >>> left = spark.createDataFrame(
        ...     pd.DataFrame(
        ...         {
        ...             "a": [0, 1, 2, 3],
        ...             "b": ["a", "b", "c", "d"],
        ...             "c": [1, 1, 1, 1],
        ...             "d": ["2", "2", "2", "2"],
        ...             "n": ["a", "b", "c", "d"],
        ...         }
        ...     )
        ... )
        ... right = left.where("a in ("1", "2")")
        >>>
        >>> # Check
        ```{.py .python linenums="1" title="Check"}
        >>> left.show()
        >>> right.show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+---+
        | a | b | c | d | n |
        +---+---+---+---+---+
        | 1 | a | 1 | 2 | a |
        | 2 | b | 1 | 2 | b |
        | 3 | c | 1 | 2 | c |
        | 4 | d | 1 | 2 | d |
        +---+---+---+---+---+
        ```
        ```{.txt .text title="Terminal"}
        +---+---+---+---+---+
        | a | b | c | d | n |
        +---+---+---+---+---+
        | 1 | a | 1 | 2 | a |
        | 2 | b | 1 | 2 | b |
        +---+---+---+---+---+
        ```
        </div>

        ```{.py .python linenums="1" title="Example 1: Single column"}
        >>> drop_matching_rows(
        ...     left_table=left,
        ...     right_table=right,
        ...     on_keys=["a"],
        ... ).show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+---+
        | a | b | c | d | n |
        +---+---+---+---+---+
        | 3 | c | 1 | 2 | c |
        | 4 | d | 1 | 2 | d |
        +---+---+---+---+---+
        ```
        !!! success "Conclusion: Successfully removed the records from the `left_table` which are existing on the `right_table`."
        </div>

        ```{.py .python linenums="1" title="Example 2: Single column as string"}
        >>> left.transform(
        ...     drop_matching_rows,
        ...     right_table=right,
        ...     on_keys="a",
        ... ).show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+---+
        | a | b | c | d | n |
        +---+---+---+---+---+
        | 3 | c | 1 | 2 | c |
        | 4 | d | 1 | 2 | d |
        +---+---+---+---+---+
        ```
        !!! success "Conclusion: Successfully removed the records from the `left_table` which are existing on the `right_table`."
        </div>

        ```{.py .python linenums="1" title="Example 3: Multiple key columns"}
        >>> drop_matching_rows(
        ...     left_table=left,
        ...     right_table=right,
        ...     on_keys=["a", "b"],
        ... ).show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+---+
        | a | b | c | d | n |
        +---+---+---+---+---+
        | 3 | c | 1 | 2 | c |
        | 4 | d | 1 | 2 | d |
        +---+---+---+---+---+
        ```
        !!! success "Conclusion: Successfully removed the records from the `left_table` which are existing on the `right_table`."
        </div>

        ```{.py .python linenums="1" title="Example 4: Including `where` clause"}
        >>> drop_matching_rows(
        ...     left_table=left,
        ...     right_table=right,
        ...     on_keys=["a"],
        ...     where_clause="n <> 'd'",
        ... ).show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+---+
        | a | b | c | d | n |
        +---+---+---+---+---+
        | 3 | c | 1 | 2 | c |
        +---+---+---+---+---+
        ```
        !!! success "Conclusion: Successfully removed the records from the `left_table` which are existing on the `right_table` and matched the `where` clause."
        </div>

        ```{.py .python linenums="1" title="Example 5: Part of pipe"}
        >>> left.transform(
        ...     func=drop_matching_rows,
        ...     right_table=right,
        ...     on_keys=["a"],
        ... ).show()
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        +---+---+---+---+---+
        | a | b | c | d | n |
        +---+---+---+---+---+
        | 3 | c | 1 | 2 | c |
        | 4 | d | 1 | 2 | d |
        +---+---+---+---+---+
        ```
        !!! success "Conclusion: Successfully removed the records from the `left_table` which are existing on the `right_table`."
        </div>

        ```{.py .python linenums="1" title="Example 6: Invalid column"}
        >>> drop_matching_rows(
        ...     left_table=left,
        ...     right_table=right,
        ...     on_keys=["f"],
        ... )
        ```
        <div class="result" markdown>
        ```{.txt .text title="Terminal"}
        ColumnDoesNotExistError: Columns ['f'] do not exist in the DataFrame.
        Try one of: ["a", "b", "c", "d", "n"]
        ```
        !!! failure "Conclusion: Columns do not exist."
        </div>

    ??? info "Notes"
        - The `on_keys` parameter can be a single string or a list of strings. This is to allow for multiple columns to be used as the matching keys.
        - The `where_clause` parameter is optional. If specified, then only the records which match the condition will be kept on the `left_table`. It is applied after the join. If not specified, then all records which are existing on the `right_table` will be removed from the `left_table`.

    ??? tip "See Also"
        - [`assert_columns_exists()`][toolbox_pyspark.checks.assert_columns_exists]
    """
    on_keys = [on_keys] if is_type(on_keys, str) else on_keys
    assert_columns_exists(left_table, on_keys, False)
    assert_columns_exists(right_table, on_keys, False)
    return (
        left_table.alias("left")
        .join(right_table.alias("right"), on=on_keys, how=join_type)
        .where("1=1" if where_clause is None else where_clause)
        .select([f"left.{col}" for col in left_table.columns])
    )



# ============================================================================ #
#                                                                              #
#     Title   : Constants                                                      #
#     Purpose : Hold the definitions of all constant values used across the    #
#               package.                                                       #
#                                                                              #
# ============================================================================ #


# ---------------------------------------------------------------------------- #
#                                                                              #
#     Overview                                                              ####
#                                                                              #
# ---------------------------------------------------------------------------- #


# ---------------------------------------------------------------------------- #
#  Description                                                              ####
# ---------------------------------------------------------------------------- #


"""
!!! note "Summary"
    The `constants` module is used to hold the definitions of all constant values used across the package.
"""


# ---------------------------------------------------------------------------- #
#                                                                              #
#     Setup                                                                 ####
#                                                                              #
# ---------------------------------------------------------------------------- #


# ---------------------------------------------------------------------------- #
#  Imports                                                                  ####
# ---------------------------------------------------------------------------- #


# ## Python StdLib Imports ----
from functools import partial
from typing import Literal, Union, get_args
from warnings import warn

# ## Python Third Party Imports ----
from pyspark.sql import types as T
from pyspark.sql.types import _all_atomic_types as pyspark_atomic_types
from toolbox_python.collection_types import str_list, str_set

# ## Local First Party Imports ----
from toolbox_pyspark.utils.whitespaces import WhitespaceCharacters, WhitespaceChatacter


# ---------------------------------------------------------------------------- #
#  Exports                                                                  ####
# ---------------------------------------------------------------------------- #


__all__: str_list = [
    "ALL_WHITESPACE_CHARACTERS",
    "WHITESPACE_CHARACTERS",
    "VALID_PYSPARK_TYPES",
    "VALID_PYSPARK_TYPE_NAMES",
    "ALL_PYSPARK_TYPES",
    "VALID_PYAPARK_JOIN_TYPES",
    "ALL_PYSPARK_JOIN_TYPES",
    "LITERAL_PANDAS_DATAFRAME_NAMES",
    "LITERAL_PYSPARK_DATAFRAME_NAMES",
    "LITERAL_NUMPY_ARRAY_NAMES",
    "LITERAL_LIST_OBJECT_NAMES",
    "VALID_PANDAS_DATAFRAME_NAMES",
    "VALID_PYSPARK_DATAFRAME_NAMES",
    "VALID_NUMPY_ARRAY_NAMES",
    "VALID_LIST_OBJECT_NAMES",
    "VALID_DATAFRAME_NAMES",
    "_DEFAULT_DEPRECATION_WARNING_CLASS",
    "_DEFAULT_DEPRECATION_WARNING",
]


# ---------------------------------------------------------------------------- #
#                                                                              #
#     Constants                                                             ####
#                                                                              #
# ---------------------------------------------------------------------------- #


# ---------------------------------------------------------------------------- #
#  White Spaces                                                             ####
# ---------------------------------------------------------------------------- #


# For full list of characters: https://en.wikipedia.org/wiki/Whitespace_character
# in the below tuples: ('name','unicode','ascii')
ALL_WHITESPACE_CHARACTERS: list[tuple[str, str, int]] = [
    ("character tabulation", "U+0009", 9),
    ("line feed", "U+000A", 10),
    ("line tabulation", "U+000B", 11),
    ("form feed", "U+000C", 12),
    ("carriage return", "U+000D", 13),
    ("space", "U+0020", 32),
    ("next line", "U+0085", 133),
    ("no-break space", "U+00A0", 160),
    ("ogham space mark", "U+1680", 5760),
    ("en quad", "U+2000", 8192),
    ("em quad", "U+2001", 8193),
    ("en space", "U+2002", 8194),
    ("em space", "U+2003", 8195),
    ("three-per-em space", "U+2004", 8196),
    ("four-per-em space", "U+2005", 8197),
    ("six-per-em space", "U+2006", 8198),
    ("figure space", "U+2007", 8199),
    ("punctuation space", "U+2008", 8200),
    ("thin space", "U+2009", 8201),
    ("hair space", "U+200A", 8202),
    ("line separator", "U+2028", 8232),
    ("paragraph separator", "U+2029", 8233),
    ("narrow no-break space", "U+202F", 8239),
    ("medium mathematical space", "U+205F", 8287),
    ("ideographic space", "U+3000", 12288),
    ("mongolian vowel separator", "U+180E", 6158),
    ("zero width space", "U+200B", 8203),
    ("zero width non-joiner", "U+200C", 8204),
    ("zero width joiner", "U+200D", 8205),
    ("word joiner", "U+2060", 8288),
    ("zero width non-breaking space", "U+FEFF", 65279),
]

WHITESPACE_CHARACTERS = WhitespaceCharacters(
    [
        WhitespaceChatacter(name, unicode, ascii)
        for name, unicode, ascii in ALL_WHITESPACE_CHARACTERS
    ]
)


# ---------------------------------------------------------------------------- #
#  PySpark Types                                                            ####
# ---------------------------------------------------------------------------- #


# For a full list of valid types, see: https://spark.apache.org/docs/latest/sql-ref-datatypes.html
VALID_PYSPARK_TYPES = list(pyspark_atomic_types.values())
VALID_PYSPARK_TYPE_NAMES: str_list = sorted(
    list(pyspark_atomic_types.keys()) + ["str", "int", "bool", "datetime"]
)
ALL_PYSPARK_TYPES = Union[
    T.DataType,
    T.NullType,
    T.CharType,
    T.StringType,
    T.VarcharType,
    T.BinaryType,
    T.BooleanType,
    T.DateType,
    T.TimestampType,
    T.TimestampNTZType,
    T.DecimalType,
    T.DoubleType,
    T.FloatType,
    T.ByteType,
    T.IntegerType,
    T.LongType,
    T.DayTimeIntervalType,
    T.YearMonthIntervalType,
    T.ShortType,
    T.ArrayType,
    T.MapType,
    T.StructType,
]

VALID_PYAPARK_JOIN_TYPES = Literal[
    "inner",
    "cross",
    "outer",
    "full",
    "fullouter",
    "full_outer",
    "left",
    "leftouter",
    "left_outer",
    "right",
    "rightouter",
    "right_outer",
    "semi",
    "leftsemi",
    "left_semi",
    "anti",
    "leftanti",
    "left_anti",
]
ALL_PYSPARK_JOIN_TYPES = set(get_args(VALID_PYAPARK_JOIN_TYPES))


# ---------------------------------------------------------------------------- #
#  DataFrames                                                               ####
# ---------------------------------------------------------------------------- #


LITERAL_PANDAS_DATAFRAME_NAMES = Literal[
    "pandas.DataFrame",
    "pandas",
    "pd.DataFrame",
    "pd.df",
    "pddf",
    "pdDataFrame",
    "pdDF",
    "pd",
]

LITERAL_PYSPARK_DATAFRAME_NAMES = Literal[
    "spark.DataFrame",
    "pyspark.DataFrame",
    "pyspark",
    "spark",
    "ps.DataFrame",
    "ps.df",
    "psdf",
    "psDataFrame",
    "psDF",
    "ps",
]

LITERAL_NUMPY_ARRAY_NAMES = Literal[
    "numpy.array",
    "np.array",
    "np",
    "numpy",
    "nparr",
    "npa",
    "np.arr",
    "np.a",
]

LITERAL_LIST_OBJECT_NAMES = Literal["list", "lst", "l", "flat_list", "flatten_list"]

VALID_PANDAS_DATAFRAME_NAMES = set(get_args(LITERAL_PANDAS_DATAFRAME_NAMES))
VALID_PYSPARK_DATAFRAME_NAMES = set(get_args(LITERAL_PYSPARK_DATAFRAME_NAMES))
VALID_NUMPY_ARRAY_NAMES = set(get_args(LITERAL_NUMPY_ARRAY_NAMES))
VALID_LIST_OBJECT_NAMES = set(get_args(LITERAL_LIST_OBJECT_NAMES))

VALID_DATAFRAME_NAMES: str_set = VALID_PANDAS_DATAFRAME_NAMES.union(
    VALID_PYSPARK_DATAFRAME_NAMES
)


# ---------------------------------------------------------------------------- #
#  Other                                                                    ####
# ---------------------------------------------------------------------------- #


_DEFAULT_DEPRECATION_WARNING_CLASS = DeprecationWarning
_DEFAULT_DEPRECATION_WARNING = partial(
    warn, category=_DEFAULT_DEPRECATION_WARNING_CLASS
)
